from abc import ABC
from typing import TYPE_CHECKING, Any

from slixmpp import JID

from ...util.types import AnySession, MessageOrPresenceTypeVar

if TYPE_CHECKING:
    from ...util.types import AnyGateway


class Base(ABC):
    session: AnySession = NotImplemented
    xmpp: "AnyGateway" = NotImplemented

    jid: JID = NotImplemented
    name: str | None = NotImplemented

    @property
    def user_jid(self) -> JID:
        return self.session.user_jid

    @property
    def user_pk(self) -> int:
        return self.session.user_pk


class BaseSender(Base):
    is_participant: bool = NotImplemented

    def _send(
        self,
        stanza: MessageOrPresenceTypeVar,
        **send_kwargs: Any,  # noqa:ANN401
    ) -> MessageOrPresenceTypeVar:
        raise NotImplementedError



"""
XEP-0184 Delivery Receipts

The corresponding slixmpp module is a bit too rigid, this is our implementation
to selectively choose when we send delivery receipts
"""

from typing import Generic, TypeVar

from slixmpp import JID, Message
from slixmpp.types import MessageTypes

GatewayType = TypeVar("GatewayType")


class DeliveryReceipt(Generic[GatewayType]):
    xmpp: GatewayType

    def __init__(self, xmpp: GatewayType) -> None:
        self.xmpp = xmpp

    def ack(self, msg: Message) -> None:
        """
        Send a XEP-0184 (delivery receipt) in response to a message,
        if appropriate.

        :param msg:
        """
        if not self.requires_receipt(msg):
            return
        ack = self.make_ack(msg["id"], msg["to"], msg["from"].bare, msg["type"])
        ack.send()

    def make_ack(
        self, msg_id: str, mfrom: JID, mto: JID, mtype: MessageTypes = "chat"
    ) -> Message:
        ack = self.xmpp.Message()
        ack["type"] = mtype
        ack["to"] = mto
        ack["from"] = mfrom
        ack["receipt"] = msg_id
        return ack

    def requires_receipt(self, msg: Message) -> bool:
        """
        Check if a message is eligible for a delivery receipt.

        :param msg:
        :return:
        """
        return (
            msg["request_receipt"]
            and msg["type"] in self.xmpp.plugin["xep_0184"].ack_types
            and not msg["receipt"]
        )



import unittest.mock
from collections.abc import Callable, Iterator

import pytest
from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError

from slidge import (  # type:ignore[attr-defined]
    BaseGateway,
    BaseSession,
    LegacyBookmarks,
    LegacyContact,
    LegacyRoster,
)
from slidge.command import FormField
from slidge.command.base import (
    ConfirmationRecipient,
    ContactCommand,
    FormRecipient,
    MUCCommand,
)
from slidge.group import LegacyMUC, LegacyParticipant
from slidge.util.test import SlidgeTest  # type:ignore[attr-defined]


class Session(BaseSession["Contact | LegacyMUC[LegacyParticipant]"]):
    async def login(self) -> None:
        pass


class Gateway(BaseGateway):
    COMPONENT_NAME = "Megacorp Chat XMPP Gateway"
    GROUPS = True


class Contact(LegacyContact):
    pass


class Roster(LegacyRoster[Contact]):
    async def jid_username_to_legacy_id(self, jid_username: str) -> str:
        if jid_username.startswith("room-"):
            raise XMPPError("item-not-found", "This is a MUC JID")
        return jid_username

    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        if legacy_id.startswith("room-"):
            raise XMPPError("item-not-found", "This is a MUC JID")
        return legacy_id


class Bookmarks(LegacyBookmarks[LegacyMUC[LegacyParticipant]]):
    async def jid_username_to_legacy_id(self, jid_username: str) -> str:
        if not jid_username.startswith("room-"):
            raise XMPPError("item-not-found", "This is a contact JID")
        return jid_username

    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        if not legacy_id.startswith("room-"):
            raise XMPPError("item-not-found", "This is a contact JID")
        return legacy_id

    async def fill(self) -> None:
        pass


class ContactCommand1(ContactCommand[Contact]):
    NAME = "contact command #1"
    NODE = CHAT_COMMAND = "command1"
    HELP = "Do stuff"

    @staticmethod
    async def run(contact: Contact, *args: str) -> str:
        return f"it worked 1! {contact.legacy_id}"


class ContactCommand2(ContactCommand[Contact]):
    NAME = "contact command #2"
    NODE = CHAT_COMMAND = "command2"
    HELP = "Do some stuff"

    @staticmethod
    async def run(contact: Contact, *args: str) -> FormRecipient[Contact] | str:
        if args:
            return f"Single arg: {args[0]}"

        return FormRecipient[Contact](
            title="a title",
            instructions="some instructions",
            fields=[FormField(var="a var", label="a label", value=contact.legacy_id)],
            handler=ContactCommand2.confirm,
        )

    @staticmethod
    async def confirm(
        contact: Contact, form: dict[str, str | JID | bool]
    ) -> ConfirmationRecipient[Contact]:
        return ConfirmationRecipient(
            prompt=str(form["a var"]), handler=ContactCommand2.finish
        )

    @staticmethod
    async def finish(contact: Contact) -> str:
        return "yay!"


class MUCCommand1(MUCCommand[LegacyMUC[LegacyParticipant]]):
    NAME = "room command #1"
    NODE = CHAT_COMMAND = "command1"
    HELP = "Do stuff"

    @staticmethod
    async def run(muc: LegacyMUC[LegacyParticipant], *args: str) -> str:
        return f"it worked 1! {muc.legacy_id}"


class MUCCommand2(MUCCommand[LegacyMUC[LegacyParticipant]]):
    NAME = "room command #2"
    NODE = CHAT_COMMAND = "command2"
    HELP = "Do some stuff"

    @staticmethod
    async def run(
        muc: LegacyMUC[LegacyParticipant], *args: str
    ) -> FormRecipient[LegacyMUC[LegacyParticipant]] | str:
        if args:
            return f"Single arg: {args[0]}"

        return FormRecipient(
            title="a title",
            instructions="some instructions",
            fields=[FormField(var="a var", label="a label", value=muc.legacy_id)],
            handler=MUCCommand2.confirm,
        )

    @staticmethod
    async def confirm(
        muc: LegacyMUC[LegacyParticipant],
        form: dict[str, str | JID | bool],
    ) -> ConfirmationRecipient[LegacyMUC[LegacyParticipant]]:
        return ConfirmationRecipient(
            prompt=str(form["a var"]), handler=MUCCommand2.finish
        )

    @staticmethod
    async def finish(muc: LegacyMUC[LegacyParticipant]) -> str:
        return "yay!"


class _AdhocSetup(SlidgeTest):  # type:ignore[misc]
    plugin = globals()

    def setUp(self) -> None:
        super().setUp()
        self._add_slidge_user()
        self.__patch = unittest.mock.patch(
            "slixmpp.plugins.xep_0050.adhoc.XEP_0050.new_session", return_value="sid"
        )
        self.__patch.start()

    def tearDown(self) -> None:
        super().tearDown()
        self.__patch.stop()


class _AdhocTests:
    _test_command_jid: str
    _test_command_kind: str
    _test_command_legacy_id: str

    recv: Callable[[str], None]
    send: Callable[[str], None]
    next_sent: Callable[[], Iq]

    def test_info(self) -> None:
        self.recv(  # language=XML
            f"""
            <iq type='get'
                to='{self._test_command_jid}'
                from='romeo@shakespeare.lit/client'>
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        iq = self.next_sent()
        assert "http://jabber.org/protocol/commands" in iq["disco_info"]["features"]

    def test_items(self) -> None:
        self.recv(  # language=XML
            f"""
            <iq type='get'
                to='{self._test_command_jid}'
                from='romeo@shakespeare.lit/client'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type="result"
                to="romeo@shakespeare.lit/client"
                from="{self._test_command_jid}"
                id="1">
              <query xmlns="http://jabber.org/protocol/disco#items">
                <item jid="{self._test_command_jid}"
                      node="command1"
                      name="{self._test_command_kind} command #1" />
                <item jid="{self._test_command_jid}"
                      node="command2"
                      name="{self._test_command_kind} command #2" />
              </query>
            </iq>
            """
        )

    def test_command_info(self) -> None:
        self.recv(  # language=XML
            f"""
            <iq type='get'
                to='{self._test_command_jid}'
                from='romeo@shakespeare.lit/client'>
              <query xmlns='http://jabber.org/protocol/disco#info'
                     node='command1'
                     name='{self._test_command_kind} command #1' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type="result"
                to="romeo@shakespeare.lit/client"
                from="{self._test_command_jid}"
                id="1">
              <query xmlns='http://jabber.org/protocol/disco#info'
                     node='command1'>
                <identity name='{self._test_command_kind} command #1'
                          category='automation'
                          type='command-node' />
                <feature var='http://jabber.org/protocol/commands' />
                <feature var='jabber:x:data' />
              </query>
            </iq>
            """
        )

    def test_execute_basic(self) -> None:
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='romeo@shakespeare.lit/client'
                to='{self._test_command_jid}'
                id='exec1'>
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type="result"
                from="{self._test_command_jid}"
                to="romeo@shakespeare.lit/client"
                id="exec1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="sid"
                       status="completed">
                <note type="info">it worked 1! {self._test_command_legacy_id}</note>
              </command>
            </iq>
            """
        )

    def test_execute_multistage(self) -> None:
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='romeo@shakespeare.lit/client'
                to='{self._test_command_jid}'
                id='exec1'>
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command2'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type="result"
                from="{self._test_command_jid}"
                to="romeo@shakespeare.lit/client"
                id="exec1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="sid"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>a title</title>
                  <instructions>some instructions</instructions>
                  <field var="a var"
                         label="a label"
                         type="text-single">
                    <value>{self._test_command_legacy_id}</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type="set"
                to="{self._test_command_jid}"
                from="romeo@shakespeare.lit/client"
                id="exec1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="sid">
                <x xmlns="jabber:x:data"
                   type="submit">
                  <field var="a var">
                    <value>YAYAYA</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type="result"
                to="romeo@shakespeare.lit/client"
                from="{self._test_command_jid}"
                id="exec1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="sid"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>YAYAYA</title>
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type="set"
                to="{self._test_command_jid}"
                from="romeo@shakespeare.lit/client"
                id="exec1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="sid">
                <x xmlns="jabber:x:data"
                   type="submit">
                  <field var="confirm">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            f"""
            <iq type="result"
                from="{self._test_command_jid}"
                to="romeo@shakespeare.lit/client"
                id="exec1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="sid"
                       status="completed">
                <note type="info">yay!</note>
              </command>
            </iq>
            """
        )


class TestContactAdhoc(_AdhocSetup, _AdhocTests):
    _test_command_jid = "juliet@aim.shakespeare.lit/slidge"
    _test_command_kind = "contact"
    _test_command_legacy_id = "juliet"


class TestMUCAdhoc(_AdhocSetup, _AdhocTests):
    _test_command_jid = "room-command@aim.shakespeare.lit"
    _test_command_kind = "room"
    _test_command_legacy_id = "room-command"


class _ChatMixin(SlidgeTest):  # type:ignore[misc]
    plugin = globals()

    def setUp(self) -> None:
        super().setUp()
        self._add_slidge_user()

    def _recv(self, body: str) -> None:
        self.recv(  # language=XML
            f"""
            <message from='romeo@shakespeare.lit/slidge'
                     to='aim.shakespeare.lit'
                     type='chat'>
              <body>{body}</body>
            </message>
            """
        )

    def _send(self, body: str) -> None:
        self.send(  # language=XML
            f"""
            <message from="aim.shakespeare.lit"
                     to="romeo@shakespeare.lit/slidge"
                     type='chat'>
              <body>{body}</body>
            </message>
            """
        )

    def _error(self, condition: str, type: str, text: str) -> None:
        el = f"<{condition} xmlns='urn:ietf:params:xml:ns:xmpp-stanzas' />"
        self.send(  # language=XML
            f"""
            <message from="aim.shakespeare.lit"
                     to="romeo@shakespeare.lit/slidge"
                     type="error">
              <error type="{type}">{el};
              <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">{text}</text></error>
            </message>
            """
        )


class TestContactChat(_ChatMixin):
    def test_execute(self) -> None:
        self._recv("contact juliet command2")
        self._send("a title")
        self._send("some instructions")
        self._send("Default: juliet")
        self._send("a label? (or 'abort')")
        self._recv("echome")
        self._send("echome")
        self._recv("yes")
        self._send("yay!")

    def test_execute_with_args(self) -> None:
        self._recv("contact juliet command2 some-arg")
        self._send("Single arg: some-arg")

    def test_execute_with_invalid_contact(self) -> None:
        self._recv("contact room-prout command2")
        self._error("item-not-found", "cancel", "This is a MUC JID")

    def test_help(self) -> None:
        self._recv("contact")
        self._send(_HELP.format(recipient="contact"))

    def test_single(self) -> None:
        self._recv("contact bad-syntax")
        self._send(_HELP.format(recipient="contact"))
        self._error(
            "bad-request",
            "modify",
            "Contact commands require at least two parameters: contact_jid_username and command_name",
        )


class TestMUCChat(_ChatMixin):
    def test_execute(self) -> None:
        self._recv("room room-command command2")
        self._send("a title")
        self._send("some instructions")
        self._send("Default: room-command")
        self._send("a label? (or 'abort')")
        self._recv("echome")
        self._send("echome")
        self._recv("yes")
        self._send("yay!")

    def test_execute_with_args(self) -> None:
        self._recv("room room-command command2 some-arg")
        self._send("Single arg: some-arg")

    def test_execute_with_invalid_contact(self) -> None:
        self._recv("room juliet command2")
        self._error("item-not-found", "cancel", "This is a contact JID")

    def test_help(self) -> None:
        self._recv("room")
        self._send(_HELP.format(recipient="room"))

    def test_single(self) -> None:
        self._recv("room bad-syntax")
        self._send(_HELP.format(recipient="room"))
        self._error(
            "bad-request",
            "modify",
            "Contact commands require at least two parameters: room_jid_username and command_name",
        )


_HELP = """Available commands:
command1 ({recipient} command #1): Do stuff
command2 ({recipient} command #2): Do some stuff"""


@pytest.fixture(scope="module", autouse=True)
def _reset_commands() -> Iterator[None]:
    yield
    LegacyContact.commands.clear()
    LegacyMUC.commands.clear()



import base64
import contextlib
import hashlib
import logging
from copy import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING
from xml.etree import ElementTree

from slixmpp import JID, Message
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0511.stanza import LinkMetadata

from ....contact.contact import LegacyContact
from ....group.room import LegacyMUC
from ....util.types import (
    AnyRecipient,
    LinkPreview,
    RecipientType,
    Reply,
    XMPPAttachment,
    XMPPAttachmentMessage,
    XMPPMessage,
    XMPPTextMessage,
)
from ....util.util import dict_to_named_tuple, remove_emoji_variation_selector_16
from ... import config
from ..util import DispatcherMixin, exceptions_to_xmpp_errors

if TYPE_CHECKING:
    from slidge.util.types import AnyGateway


@dataclass
class _IncomingAttachment:
    attachment: XMPPAttachment
    is_sticker: bool
    cid: str | None = None


class MessageContentMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: "AnyGateway") -> None:
        super().__init__(xmpp)
        xmpp.add_event_handler("legacy_message", self.on_legacy_message)
        xmpp.add_event_handler("message_retract", self.on_message_retract)
        xmpp.add_event_handler("groupchat_message", self.on_groupchat_message)
        xmpp.add_event_handler("reactions", self.on_reactions)

    async def on_groupchat_message(self, msg: Message) -> None:
        await self.on_legacy_message(msg)

    @exceptions_to_xmpp_errors
    async def on_legacy_message(self, msg: Message) -> None:
        """
        Meant to be called from :class:`BaseGateway` only.

        :param msg:
        :return:
        """
        # we MUST not use `if m["replace"]["id"]` because it adds the tag if not
        # present. this is a problem for MUC echoed messages
        if "apply_to" in msg:
            # ignore message retraction (handled by a specific method)
            # this is an old version of the protocol that depends on the
            # message fastening deprecated xep.
            return
        if "reactions" in msg:
            # ignore message reaction fallback.
            # the reaction itself is handled by self.react_from_msg().
            return
        if "retract" in msg:
            # ignore message retraction fallback.
            # the retraction itself is handled by self.on_retract
            return
        recipient, thread = await self._get_recipient_and_thread(msg)
        replace = await self.__get_replace(msg, recipient)
        if msg.xml.find(".//{*}encrypted") is not None:
            raise XMPPError(
                "bad-request", "You cannot send encrypted messages through this gateway"
            )
        body, reply = await self.__get_reply(msg, recipient)
        cid = self.__get_xhtml_sticker_cid(msg)

        if cid:
            legacy_msg_id = await self.__dispatch_bob(
                msg.get_from(),
                cid,
                recipient,
                reply=reply,
                thread=thread,
            )
        else:
            attachments = self.__get_attachments(msg)
            if len(attachments) == 1 and attachments[0].is_sticker:
                legacy_msg_id = await self.__dispatch_nonbob_sticker(
                    attachments[0],
                    recipient,
                    msg["body"],
                    reply=reply,
                    thread=thread,
                )
            else:
                legacy_msg_id = await self.__dispatch_msg(
                    replace=replace,
                    body=body,
                    attachments=tuple(a.attachment for a in attachments),
                    recipient=recipient,
                    thread=thread,
                    reply=reply,
                    msg=msg,
                )

        if isinstance(recipient, LegacyMUC):
            stanza_id = await recipient.echo(msg, legacy_msg_id)
        else:
            stanza_id = None
            self.__ack(msg)

        if not legacy_msg_id:
            return

        with self.xmpp.store.session() as orm:
            if recipient.is_group:
                self.xmpp.store.id_map.set_origin(
                    orm, recipient.stored.id, legacy_msg_id, msg.get_id()
                )
                assert stanza_id is not None
                self.xmpp.store.id_map.set_msg(
                    orm,
                    recipient.stored.id,
                    legacy_msg_id,
                    [stanza_id],
                    True,
                )
            else:
                self.xmpp.store.id_map.set_msg(
                    orm,
                    recipient.stored.id,
                    legacy_msg_id,
                    [msg.get_id()],
                    False,
                )
            if recipient.session.MESSAGE_IDS_ARE_THREAD_IDS and thread:
                self.xmpp.store.id_map.set_thread(
                    orm, recipient.stored.id, thread, legacy_msg_id, recipient.is_group
                )
            orm.commit()

    def __get_xhtml_sticker_cid(self, msg: Message) -> str | None:
        if "html" not in msg:
            return None
        body = ElementTree.fromstring("<body>" + msg["html"].get_body() + "</body>")
        p = body.findall("p")
        if p is None:
            return None
        if len(p) != 1:
            return None
        if p[0].text is not None and p[0].text.strip():
            return None

        images = p[0].findall("img")
        if len(images) != 1:
            return None
        # no text, single img ⇒ this is a sticker
        # other cases should be interpreted as "custom emojis" in text
        src = images[0].get("src")
        if src is None:
            return None
        if src.startswith("cid:"):
            return src.removeprefix("cid:")
        return None

    async def __get_replace(
        self,
        msg: Message,
        recipient: RecipientType,
    ) -> str | None:
        if "replace" not in msg or "id" not in msg["replace"]:
            return None
        return (
            self._xmpp_msg_id_to_legacy(msg["replace"]["id"], recipient, True) or None
        )

    def __get_attachments(self, msg: Message) -> list[_IncomingAttachment]:
        is_sticker = "sticker" in msg

        if (
            "sfs" in msg
            and "sources" in msg["sfs"]
            and "url-data" in msg["sfs"]["sources"]
            and "target" in msg["sfs"]["sources"]["url-data"]
        ):
            # TODO: support "attach source in later message", cf https://xmpp.org/extensions/xep-0447.html#example-5
            # TODO: support for other sources than URL
            # TODO: support for multiattachments in single message.
            #       What do we do if is_sticker and multiple files?
            attachment = _IncomingAttachment(
                XMPPAttachment(url=msg["sfs"]["sources"]["url-data"]["target"]),
                is_sticker=is_sticker,
            )
            if "file" in msg["sfs"]:
                attachment.attachment.content_type = (
                    msg["sfs"]["file"]["media-type"] or None
                )
                if "hash" in msg["sfs"]["file"]:
                    algo = msg["sfs"]["file"]["hash"]["algo"]
                    h = msg["sfs"]["file"]["hash"]["value"]
                    attachment.cid = f"{algo}+{h}" if algo and h else None
            return [attachment]

        if "oob" in msg:
            return [
                _IncomingAttachment(
                    XMPPAttachment(url=msg["oob"]["url"]), is_sticker=is_sticker
                )
            ]

        if (
            "reference" in msg
            and "sims" in msg["reference"]
            and "sources" in msg["reference"]["sims"]
        ):
            for source in msg["reference"]["sims"]["sources"]["substanzas"]:
                if source["uri"].startswith("http"):
                    attachment = _IncomingAttachment(
                        XMPPAttachment(url=source["uri"]), is_sticker=is_sticker
                    )
                    break
            else:
                return []
            if "file" in msg["reference"]["sims"]:
                attachment.attachment.content_type = msg["media-type"] or None
            return [attachment]

        return []

    async def __dispatch_msg(
        self,
        *,
        replace: str | None,
        reply: Reply | None,
        body: str | None,
        attachments: tuple[XMPPAttachment, ...],
        recipient: AnyRecipient,
        thread: str | None,
        msg: Message,
    ) -> str | None:
        if replace:
            if body is not None:
                body = body.strip()
            if not body and not attachments and recipient.RETRACTION:
                await recipient.on_retract(replace, thread=thread)
                return None
            if not recipient.CORRECTION:
                if recipient.RETRACTION:
                    await recipient.on_retract(replace, thread=thread)
                    replace = None
                elif body:
                    body = "Correction:\n" + body

        if not any([attachments, body]):
            log.debug(
                "Ignoring msg with id '%s' because no body or attachments found",
                msg.get_id(),
            )
            return None
        link_previews = (
            parse_link_previews(msg["link_metadatas"]) if "link_metadata" in msg else ()
        )

        for attachment in attachments:
            if not body:
                body = None
                break
            body = body.replace(attachment.url, "").strip()

        if recipient.is_group:
            mentions = tuple(await recipient.parse_mentions(body))
        else:
            mentions = ()

        if attachments:
            xmpp_msg: XMPPMessage = XMPPAttachmentMessage(
                body=body,
                attachments=attachments,
                reply=reply,
                thread=thread,
                link_previews=link_previews,
                mentions=mentions,
                replace=replace,
            )
        else:
            xmpp_msg = XMPPTextMessage(
                body=body,
                attachments=attachments,
                reply=reply,
                thread=thread,
                link_previews=link_previews,
                mentions=mentions,
                replace=replace,
            )

        return await recipient.on_message(xmpp_msg)

    @exceptions_to_xmpp_errors
    async def on_message_retract(self, msg: Message) -> None:
        recipient, thread = await self._get_recipient_and_thread(msg)
        if not recipient.RETRACTION:
            raise XMPPError(
                "bad-request",
                "This legacy service does not support message retraction.",
            )
        xmpp_id: str = msg["retract"]["id"]
        legacy_id = self._xmpp_msg_id_to_legacy(xmpp_id, recipient, origin=True)
        await recipient.on_retract(legacy_id, thread=thread)
        if isinstance(recipient, LegacyMUC):
            await recipient.echo(msg, None)
        self.__ack(msg)

    @exceptions_to_xmpp_errors
    async def on_reactions(self, msg: Message) -> None:
        recipient, thread = await self._get_recipient_and_thread(msg)
        react_to: str = msg["reactions"]["id"]

        special_msg = recipient.session.SPECIAL_MSG_ID_PREFIX and react_to.startswith(
            recipient.session.SPECIAL_MSG_ID_PREFIX
        )

        if special_msg:
            legacy_id = react_to
        else:
            legacy_id = self._xmpp_msg_id_to_legacy(react_to, recipient)

        if not legacy_id:
            log.debug("Ignored reaction from user")
            raise XMPPError(
                "internal-server-error",
                "Could not convert the XMPP msg ID to a legacy ID",
            )

        emojis = [
            remove_emoji_variation_selector_16(r["value"]) for r in msg["reactions"]
        ]
        error_msg = None
        recipient = recipient

        if not special_msg:
            if recipient.REACTIONS_SINGLE_EMOJI and len(emojis) > 1:
                error_msg = "Maximum 1 emoji/message"

            if (
                not error_msg
                and (subset := await recipient.available_emojis(legacy_id))
                and not set(emojis).issubset(subset)
            ):
                error_msg = (
                    f"You can only react with the following emojis: {''.join(subset)}"
                )

        if error_msg:
            recipient.session.send_gateway_message(error_msg)
            if not isinstance(recipient, LegacyMUC):
                # no need to carbon for groups, we just don't echo the stanza
                recipient.react(legacy_id, carbon=True)
            await recipient.on_react(legacy_id, [], thread=thread)
            raise XMPPError(
                "policy-violation",
                text=error_msg,
                clear=False,
            )

        await recipient.on_react(legacy_id, emojis, thread=thread)
        if isinstance(recipient, LegacyMUC):
            await recipient.echo(msg, None)
        else:
            self.__ack(msg)

        with self.xmpp.store.session() as orm:
            multi = self.xmpp.store.id_map.get_xmpp(
                orm, recipient.stored.id, legacy_id, recipient.is_group
            )
        if not multi:
            return
        multi = [m for m in multi if react_to != m]

        if isinstance(recipient, LegacyMUC):
            for xmpp_id in multi:
                mc = copy(msg)
                mc["reactions"]["id"] = xmpp_id
                await recipient.echo(mc)
        elif isinstance(recipient, LegacyContact):
            for xmpp_id in multi:
                recipient.react(legacy_id, emojis, xmpp_id=xmpp_id, carbon=True)

    def __ack(self, msg: Message) -> None:
        if not self.xmpp.PROPER_RECEIPTS:
            self.xmpp.delivery_receipt.ack(msg)

    async def __get_reply(
        self, msg: Message, recipient: AnyRecipient
    ) -> tuple[str, Reply | None]:
        if "reply" not in msg:
            return msg["body"], None

        session = recipient.session

        try:
            reply_to_msg_id = self._xmpp_msg_id_to_legacy(msg["reply"]["id"], recipient)
        except XMPPError:
            session.log.debug(
                "Could not determine reply-to legacy msg ID, sending quote instead."
            )
            return redact_url(msg["body"]), None

        reply_to_jid = JID(msg["reply"]["to"])
        reply_to = None
        if msg["type"] == "chat":
            if reply_to_jid.bare != session.user_jid.bare:
                with contextlib.suppress(XMPPError):
                    reply_to = await session.contacts.by_jid(reply_to_jid)
        elif msg["type"] == "groupchat":
            nick = reply_to_jid.resource
            try:
                muc = await session.bookmarks.by_jid(reply_to_jid)
            except XMPPError:
                pass
            else:
                if nick == muc.user_nick:
                    reply_to = await muc.get_user_participant()
                elif not nick:
                    reply_to = muc.get_system_participant()
                else:
                    reply_to = await muc.get_participant(nick, store=False)

        if "fallback" in msg and (
            isinstance(recipient, LegacyMUC) or recipient.REPLIES
        ):
            text = msg["fallback"].get_stripped_body(self.xmpp["xep_0461"].namespace)
            try:
                reply_fallback = redact_url(msg["reply"].get_fallback_body())
            except AttributeError:
                reply_fallback = None
        else:
            text = msg["body"]
            reply_fallback = None

        return text, Reply(reply_to_msg_id, reply_to, reply_fallback)

    async def __dispatch_nonbob_sticker(
        self,
        attachment: _IncomingAttachment,
        recipient: AnyRecipient,
        fallback: str,
        reply: Reply | None = None,
        thread: str | None = None,
    ) -> str | None:
        if attachment.cid:
            with self.xmpp.store.session() as orm:
                sticker = self.xmpp.store.bob.get_sticker(orm, attachment.cid)
        else:
            sticker = None
        if sticker is None:
            async with attachment.attachment.get() as response:
                response.raise_for_status()
                data = await response.read()
            cid = "sha256+" + hashlib.sha256(data).hexdigest()
            with self.xmpp.store.session() as orm:
                sticker = self.xmpp.store.bob.set_sticker(
                    orm, cid, data, attachment.attachment.content_type
                )
                orm.commit()
            if sticker is None:
                return await recipient.on_message(
                    XMPPTextMessage(
                        body="\n".join([attachment.attachment.url, fallback]),
                        thread=thread,
                        reply=reply,
                    )
                )
        sticker.reply = reply
        sticker.thread = thread
        return await recipient.on_sticker(sticker)

    async def __dispatch_bob(
        self,
        from_: JID,
        cid: str,
        recipient: AnyRecipient,
        reply: Reply | None = None,
        thread: str | None = None,
    ) -> str | None:
        with self.xmpp.store.session() as orm:
            sticker = self.xmpp.store.bob.get_sticker(orm, cid)
        if sticker is None:
            await self.xmpp.plugin["xep_0231"].get_bob(
                from_, cid, ifrom=self.xmpp.boundjid
            )
            with self.xmpp.store.session() as orm:
                sticker = self.xmpp.store.bob.get_sticker(orm, cid)
        assert sticker is not None
        sticker.reply = reply
        sticker.thread = thread
        return await recipient.on_sticker(sticker)


def redact_url(text: str) -> str:
    needle = config.NO_UPLOAD_URL_PREFIX or config.UPLOAD_URL_PREFIX
    if not needle:
        return text
    return text.replace(needle, "")


def parse_link_previews(link_metadatas: list[LinkMetadata]) -> tuple[LinkPreview, ...]:
    result = []
    for link_metadata in link_metadatas:
        preview: LinkPreview = dict_to_named_tuple(link_metadata, LinkPreview)  # type:ignore[arg-type]
        if (
            preview.image
            and isinstance(preview.image, str)
            and preview.image.startswith("data:image/jpeg;base64,")
        ):
            try:
                image = base64.b64decode(
                    preview.image.removeprefix("data:image/jpeg;base64,")
                )
            except Exception as e:
                log.warning(
                    "Could not decode base64-encoded image: %s '%s'", e, preview.image
                )
            else:
                preview = preview._replace(image=image)
        result.append(preview)
    return tuple(result)


log = logging.getLogger(__name__)



import unittest.mock

from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError

from slidge import (
    BaseGateway,
    BaseSession,
    LegacyBookmarks,
    LegacyContact,
    LegacyMUC,
    LegacyParticipant,
)
from slidge.command.user import _get_updated_spaces
from slidge.util.test import SlidgeTest
from slidge.util.types import SpaceMetadata


class Session(BaseSession):
    async def login(self):
        pass


class Gateway(BaseGateway):
    COMPONENT_NAME = "Megacorp Chat XMPP Gateway"
    GROUPS = True
    SPACES = True


class Contact(LegacyContact):
    async def update_info(self):
        if self.legacy_id in ("programming-socks", "boring-software", "editors"):
            raise XMPPError("item-not-found", "Not a contact")


class Bookmarks(LegacyBookmarks["MUC"]):
    async def fill(self):
        for legacy_id in (
            "programming-socks",
            "boring-software",
            "editors",
            "fancy-hotels",
            "finance",
        ):
            await self.by_legacy_id(legacy_id)

    async def fetch_space_metadata(self, legacy_id: str):
        match legacy_id:
            case "dev":
                return SpaceMetadata(
                    creator_legacy_id="lead-dev",
                    owner_legacy_ids=["lead-dev", "lead-dev-friend"],
                    name="The developers corner",
                    description="Nerds welcome!",
                    member_count=666,
                )
            case "sales":
                return SpaceMetadata(
                    creator_legacy_id="ferrari-guy",
                    name="Sales team",
                )
            case "management":
                return SpaceMetadata(
                    creator_legacy_id="ceo",
                    name="The bosses lounge",
                )
            case _:
                raise XMPPError("item-not-found", "Not a space")


class MUC(LegacyMUC[LegacyParticipant]):
    async def update_info(self):
        match self.legacy_id:
            case "programming-socks":
                self.name = "Share your best programming socks deal HERE"
                self.space_legacy_id = "dev"
            case "boring-software":
                self.name = "ERP 3000"
                self.space_legacy_id = "dev"
            case "editors":
                self.name = "Settling the text editor wars"
                self.space_legacy_id = "dev"
            case "fancy-hotels":
                self.space_legacy_id = "sales"
            case "finance":
                self.space_legacy_id = "management"
            case "new-dev":
                self.space_legacy_id = "dev"


class TestSpaces(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        self._add_slidge_user("emacs-user@shakespeare.lit")

    def user_session(
        self, jid: str | JID = "emacs-user@shakespeare.lit"
    ) -> "BaseSession":
        return super().user_session(jid)

    def test_query_spaces_not_implemented(self):
        Gateway.SPACES = False
        self.recv(  # language=XML
            """
            <iq type='get'
                from='emacs-user@shakespeare.lit/cool-client'
                to='aim.shakespeare.lit'
                id='spaces-items'>
              <query xmlns='http://jabber.org/protocol/disco#items'>
                <filter xmlns='urn:xmpp:pubsub-filter:0'>
                  <x xmlns='jabber:x:data'
                     type='submit'>
                    <field var='FORM_TYPE'
                           type='hidden'>
                      <value>urn:xmpp:pubsub-filter:0</value>
                    </field>
                    <field type='list-multi'
                           var='included-types'>
                      <value>urn:xmpp:spaces:0</value>
                    </field>
                  </x>
                </filter>
              </query>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="spaces-items">
              <error type="cancel">
                <feature-not-implemented xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">This gateway does not support XEP-0503</text>
              </error>
            </iq>
            """
        )
        Gateway.SPACES = True

    def test_query_spaces(self):
        self.recv(  # language=XML
            """
            <iq type='get'
                from='emacs-user@shakespeare.lit/cool-client'
                to='aim.shakespeare.lit'
                id='spaces-items'>
              <query xmlns='http://jabber.org/protocol/disco#items'>
                <filter xmlns='urn:xmpp:pubsub-filter:0'>
                  <x xmlns='jabber:x:data'
                     type='submit'>
                    <field var='FORM_TYPE'
                           type='hidden'>
                      <value>urn:xmpp:pubsub-filter:0</value>
                    </field>
                    <field type='list-multi'
                           var='included-types'>
                      <value>urn:xmpp:spaces:0</value>
                    </field>
                  </x>
                </filter>
              </query>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type='result'
                to='emacs-user@shakespeare.lit/cool-client'
                from='aim.shakespeare.lit'
                id='spaces-items'>
              <query xmlns='http://jabber.org/protocol/disco#items'>
                <item jid='aim.shakespeare.lit'
                      node='dev'
                      name='The developers corner' />
                <item jid='aim.shakespeare.lit'
                      node='sales'
                      name='Sales team' />
                <item jid='aim.shakespeare.lit'
                      node='management'
                      name='The bosses lounge' />
              </query>
            </iq>
            """
        )

    def test_query_space_metadata(self):
        self.recv(  # language=XML
            """
            <iq type='get'
                from='emacs-user@shakespeare.lit/cool-client'
                to='aim.shakespeare.lit'
                id='space-metadata'>
              <query xmlns='http://jabber.org/protocol/disco#info'
                     node='dev' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type='result'
                from='aim.shakespeare.lit'
                to='emacs-user@shakespeare.lit/cool-client'
                id='space-metadata'>
              <query xmlns='http://jabber.org/protocol/disco#info'
                     node='dev'>
                <identity category='pubsub'
                          type='leaf' />
                <feature var='http://jabber.org/protocol/pubsub' />
                <feature var='http://jabber.org/protocol/pubsub#meta-data' />
                <feature var='http://jabber.org/protocol/pubsub#item-ids' />
                <feature var='http://jabber.org/protocol/pubsub#manage-subscriptions' />
                <feature var='http://jabber.org/protocol/pubsub#modify-affiliations' />
                <feature var='http://jabber.org/protocol/pubsub#outcast-affiliation' />
                <feature var='http://jabber.org/protocol/pubsub#retract-items' />
                <feature var='http://jabber.org/protocol/pubsub#retrieve-affiliations' />
                <feature var='http://jabber.org/protocol/pubsub#retrieve-items' />
                <feature var='http://jabber.org/protocol/pubsub#retrieve-subscriptions' />
                <feature var='http://jabber.org/protocol/pubsub#subscribe' />
                <feature var='http://jabber.org/protocol/pubsub#subscription-notifications' />
                <x xmlns='jabber:x:data'
                   type='result'>
                  <field var='FORM_TYPE'
                         type='hidden'>
                    <value>http://jabber.org/protocol/pubsub#meta-data</value>
                  </field>
                  <field var='pubsub#type'>
                    <value>urn:xmpp:spaces:0</value>
                  </field>
                  <field var='pubsub#creator'
                         type='jid-single'>
                    <value>lead-dev@aim.shakespeare.lit</value>
                  </field>
                  <field var='pubsub#title'>
                    <value>The developers corner</value>
                  </field>
                  <field var='pubsub#owner'
                         type='jid-multi'>
                    <value>lead-dev@aim.shakespeare.lit</value>
                    <value>lead-dev-friend@aim.shakespeare.lit</value>
                  </field>
                  <field var='pubsub#num_subscribers'>
                    <value>666</value>
                  </field>
                  <field var='pubsub#description'>
                    <value>Nerds welcome!</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )

    def test_get_all_items(self):
        self.recv(  # language=XML
            """
            <iq type='get'
                from='emacs-user@shakespeare.lit/cool-client'
                to='aim.shakespeare.lit'
                id='space-items'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='dev' />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type='result'
                from='aim.shakespeare.lit'
                to='emacs-user@shakespeare.lit/cool-client'
                id='space-items'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='dev'>
                  <item id='boring-software@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1'
                                name='ERP 3000' />
                  </item>
                  <item id='programming-socks@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1'
                                name='Share your best programming socks deal HERE'></conference>
                  </item>
                  <item id='editors@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1'
                                name='Settling the text editor wars'></conference>
                  </item>
                </items>
              </pubsub>
            </iq>
            """,
            use_values=False,
        )

    def test_get_single_item(self) -> None:
        self.__recv_get_item(["programming-socks@aim.shakespeare.lit"])
        self.__send_items(
            [
                (
                    "programming-socks@aim.shakespeare.lit",
                    "Share your best programming socks deal HERE",
                )
            ]
        )

    def test_get_multiple_item(self) -> None:
        self.__recv_get_item(
            ["programming-socks@aim.shakespeare.lit", "editors@aim.shakespeare.lit"]
        )
        self.__send_items(
            [
                (
                    "programming-socks@aim.shakespeare.lit",
                    "Share your best programming socks deal HERE",
                ),
                (
                    "editors@aim.shakespeare.lit",
                    "Settling the text editor wars",
                ),
            ]
        )

    def test_get_nonexistent_item(self) -> None:
        self.__recv_get_item(["nope"])
        self.send(  # language=XML
            """
            <iq type="error"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="0">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Slidge only supports items that are MUCs.</text>
              </error>
            </iq>
            """
        )

    def test_get_nonexistent_item_multiple(self) -> None:
        self.__recv_get_item(["nope", "editors@aim.shakespeare.lit"])
        self.send(  # language=XML
            """
            <iq type="error"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="0">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Slidge only supports items that are MUCs.</text>
              </error>
            </iq>
            """
        )

    def test_get_nonexistent_item_multiple2(self) -> None:
        self.__recv_get_item(["xxx@aim.shakespeare.lit", "editors@aim.shakespeare.lit"])
        self.send(  # language=XML
            """
            <iq type="error"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="0">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Could not find items: {'xxx@aim.shakespeare.lit'}</text>
              </error>
            </iq>
            """
        )

    def __recv_get_item(self, ids: list[str], node: str = "dev") -> None:
        items = ""
        for i in ids:
            items += f"<item id='{i}' />\n"
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='emacs-user@shakespeare.lit/cool-client'
                to='aim.shakespeare.lit'
                id='0'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='{node}'>{items}</items>
              </pubsub>
            </iq>
            """
        )

    def __send_items(self, data: list[tuple[str, str]], node: str = "dev") -> None:
        items = ""
        for id_, name in data:
            items += f"<item id='{id_}'>\n<conference xmlns='urn:xmpp:bookmarks:1' name='{name}'/></item>"
        self.send(  # language=XML
            f"""
            <iq type='result'
                from='aim.shakespeare.lit'
                to='emacs-user@shakespeare.lit/cool-client'
                id='0'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='{node}'>{items}</items>
              </pubsub>
            </iq>
            """
        )

    def test_muc_advertises_membership(self):
        self.recv(  # language=XML
            """
            <iq from='emacs-user@shakespeare.lit/cool-client'
                to='programming-socks@aim.shakespeare.lit'
                type='get'>
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                from="programming-socks@aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                type="result"
                id="1">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="conference"
                          type="text"
                          name="Share your best programming socks deal HERE" />
                <feature var="http://jabber.org/protocol/muc" />
                <feature var="http://jabber.org/protocol/muc#stable_id" />
                <feature var="http://jabber.org/protocol/muc#self-ping-optimization" />
                <feature var="urn:xmpp:mam:2" />
                <feature var="urn:xmpp:mam:2#extended" />
                <feature var="urn:xmpp:sid:0" />
                <feature var="muc_persistent" />
                <feature var="vcard-temp" />
                <feature var="urn:xmpp:ping" />
                <feature var="urn:xmpp:occupant-id:0" />
                <feature var="jabber:iq:register" />
                <feature var="urn:xmpp:message-moderate:1" />
                <feature var="muc_open" />
                <feature var="muc_semianonymous" />
                <feature var="muc_public" />
                <feature var="urn:xmpp:spaces:0" />
                <feature var="http://jabber.org/protocol/commands" />
                <feature var="urn:xmpp:muc:affiliations:1" />
                <x xmlns="jabber:x:data"
                   type="result">
                  <field var="FORM_TYPE"
                         type="hidden">
                    <value>http://jabber.org/protocol/muc#roominfo</value>
                  </field>
                  <field var="muc#roomconfig_persistentroom"
                         type="boolean">
                    <value>1</value>
                  </field>
                  <field var="muc#roomconfig_changesubject"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#maxhistoryfetch">
                    <value>100</value>
                  </field>
                  <field var="muc#roominfo_subjectmod"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#roomconfig_roomname">
                    <value>Share your best programming socks deal HERE</value>
                  </field>
                  <field var="muc#roomconfig_membersonly"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#roomconfig_whois"
                         type="list-single">
                    <value>moderators</value>
                  </field>
                  <field var="muc#roomconfig_publicroom"
                         type="boolean">
                    <value>1</value>
                  </field>
                  <field var="muc#roomconfig_allowpm"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#roominfo_pubsub">
                    <value>xmpp:aim.shakespeare.lit?node=dev</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )

    def test_user_subscribe_success(self) -> None:
        self.__recv_subscription_request("dev")
        self.send(  # language=XML
            """
            <iq type="result"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="1">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscription node="dev"
                              jid="emacs-user@shakespeare.lit"
                              subscription="subscribed" />
              </pubsub>
            </iq>
            """
        )

    def test_user_subscribe_non_existing_space(self) -> None:
        self.__recv_subscription_request("non-existing")
        iq = self.next_sent()
        assert isinstance(iq, Iq)
        assert iq["type"] == "error"
        assert iq["error"]["type"] == "cancel"
        assert iq["error"]["condition"] == "item-not-found"

    def __recv_subscription_request(self, node: str) -> None:
        self.recv(  # language=XML
            f"""
            <iq to="aim.shakespeare.lit"
                from="emacs-user@shakespeare.lit/cool-client"
                type="set">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscribe jid="emacs-user@shakespeare.lit"
                           node="{node}" />
              </pubsub>
            </iq>
            """
        )

    def test_user_leave_space(self) -> None:
        with unittest.mock.patch("slidge.BaseSession.on_leave_space") as on_leave_space:
            self.recv(  # language=XML
                """
            <iq to="aim.shakespeare.lit"
                from="emacs-user@shakespeare.lit/cool-client"
                type="set">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <unsubscribe jid="emacs-user@shakespeare.lit"
                             node="dev" />
              </pubsub>
            </iq>
            """
            )
        on_leave_space.assert_awaited_once_with("dev")
        self.send(  # language=XML
            """
            <iq type="result"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="1">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscription node="dev"
                              jid="emacs-user@shakespeare.lit"
                              subscription="none" />
              </pubsub>
            </iq>
            """
        )

    def test_new_room_in_existing_space(self) -> None:
        self.run_coro(self.user_session().bookmarks.by_legacy_id("new-dev"))
        self.send(  # language=XML
            """
            <message type="headline"
                     from="aim.shakespeare.lit"
                     to="emacs-user@shakespeare.lit">
              <event xmlns='http://jabber.org/protocol/pubsub#event'>
                <items node='dev'>
                  <item id='boring-software@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1'
                                name='ERP 3000' />
                  </item>
                  <item id='programming-socks@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1'
                                name='Share your best programming socks deal HERE' />
                  </item>
                  <item id='editors@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1'
                                name='Settling the text editor wars' />
                  </item>
                  <item id='new-dev@aim.shakespeare.lit'>
                    <conference xmlns='urn:xmpp:bookmarks:1' />
                  </item>
                </items>
              </event>
            </message>
            """,
            use_values=False,
        )

    def test_get_owner_affiliations(self) -> None:
        self.recv(  # language=XML
            """
            <iq type="get"
                from="emacs-user@shakespeare.lit/cool-client"
                to="aim.shakespeare.lit"
                id="get">
              <pubsub xmlns="http://jabber.org/protocol/pubsub#owner">
                <affiliations node="dev" />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id="get">
              <error type="auth">
                <forbidden xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Slidge does not implement managing space affiliations.</text>
              </error>
            </iq>
            """,
        )

        # For slidge v0.5.0…
        # self.send(
        #     """
        #     <iq type='result'
        #         from='aim.shakespeare.lit'
        #         to='emacs-user@shakespeare.lit/cool-client'>
        #       <pubsub xmlns='http://jabber.org/protocol/pubsub#owner'>
        #         <affiliations>
        #           <affiliation jid='lead-dev@aim.shakespeare.lit' affiliation='owner'/>
        #           <affiliation jid='lead-dev-friend@aim.shakespeare.lit' affiliation='owner'/>
        #           <affiliation jid='emacs-user@shakespeare.lit' affiliation='member'/>
        #         </affiliations>
        #       </pubsub>
        #     </iq>
        #     """
        # )

    def test_get_node_affiliation(self) -> None:
        self.recv(  # language=XML
            """
            <iq type="get"
                from="emacs-user@shakespeare.lit/cool-client"
                to="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <affiliations node="dev" />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                to="emacs-user@shakespeare.lit/cool-client"
                from="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <affiliations>
                  <affiliation node='dev'
                               affiliation='subscriber' />
                </affiliations>
              </pubsub>
            </iq>
            """
        )

    def test_get_all_affiliations(self) -> None:
        self.recv(  # language=XML
            """
            <iq type="get"
                from="emacs-user@shakespeare.lit/cool-client"
                to="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <affiliations />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                to="emacs-user@shakespeare.lit/cool-client"
                from="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <affiliations>
                  <affiliation node='dev'
                               affiliation='subscriber' />
                  <affiliation node='sales'
                               affiliation='subscriber' />
                  <affiliation node='management'
                               affiliation='subscriber' />
                </affiliations>
              </pubsub>
            </iq>
            """
        )

    def test_metadata_change(self) -> None:
        self.run_coro(
            self.user_session().bookmarks.update_space_metadata(
                "dev",
                SpaceMetadata(name="The new name"),
            )
        )
        self.send(  # language=XML
            """
            <message type="headline"
                     from="aim.shakespeare.lit"
                     to="emacs-user@shakespeare.lit">
              <event xmlns='http://jabber.org/protocol/pubsub#event'>
                <configuration node='dev'>
                  <x xmlns='jabber:x:data'
                     type='result'>
                    <field var='FORM_TYPE'
                           type='hidden'>
                      <value>http://jabber.org/protocol/pubsub#node_config</value>
                    </field>
                    <field var='pubsub#title'>
                      <value>The new name</value>
                    </field>
                    <!-- MUST: identifies this node as a space -->
                    <field var='pubsub#type'>
                      <value>urn:xmpp:spaces:0</value>
                    </field>
                    <!-- MUST: subscribers are notified when items are retracted -->
                    <field var='pubsub#notify_retract'>
                      <value>true</value>
                    </field>
                    <!-- MUST: items are persisted on the node -->
                    <field var='pubsub#persist_items'>
                      <value>true</value>
                    </field>
                    <!-- MUST: items are NOT purged when the node goes offline -->
                    <field var='pubsub#purge_offline'>
                      <value>false</value>
                    </field>
                    <!-- MUST (public): anyone can subscribe/read -->
                    <field var='pubsub#access_model'>
                      <value>authorize</value>
                    </field>
                    <!-- SHOULD: subscribers notified of new subscriptions -->
                    <field var='pubsub#notify_sub'>
                      <value>true</value>
                    </field>
                    <!-- SHOULD: subscribers notified of configuration changes -->
                    <field var='pubsub#notify_config'>
                      <value>true</value>
                    </field>
                    <!-- SHOULD: subscribers notified when the node is deleted -->
                    <field var='pubsub#notify_delete'>
                      <value>true</value>
                    </field>
                    <!-- SHOULD NOT be open: only owners/publishers should publish -->
                    <field var='pubsub#publish_model'>
                      <value>publishers</value>
                    </field>
                  </x>
                </configuration>
              </event>
            </message>
            """
        )

    def test_get_owner_subscriptions(self) -> None:
        self.recv(  # language=XML
            """
            <iq type="get"
                from="emacs-user@shakespeare.lit/cool-client"
                to="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub#owner">
                <subscriptions node="dev" />
              </pubsub>
            </iq>
            """
        )

        self.send(  # language=XML
            """
            <iq type='result'
                from="aim.shakespeare.lit"
                to="emacs-user@shakespeare.lit/cool-client"
                id='0'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub#owner'>
                <subscriptions node='dev'>
                  <subscription jid='emacs-user@shakespeare.lit'
                                subscription='subscribed' />
                </subscriptions>
              </pubsub>
            </iq>
            """
        )

    def test_get_subscriptions(self) -> None:
        self.recv(  # language=XML
            """
            <iq type="get"
                from="emacs-user@shakespeare.lit/cool-client"
                to="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscriptions />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                to="emacs-user@shakespeare.lit/cool-client"
                from="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscriptions>
                  <subscription node='dev'
                                jid='emacs-user@shakespeare.lit'
                                subscription='subscribed' />
                  <subscription node='sales'
                                jid='emacs-user@shakespeare.lit'
                                subscription='subscribed' />
                  <subscription node='management'
                                jid='emacs-user@shakespeare.lit'
                                subscription='subscribed' />
                </subscriptions>
              </pubsub>
            </iq>
            """
        )

    def test_get_subscriptions_specific_node(self) -> None:
        self.recv(  # language=XML
            """
            <iq type="get"
                from="emacs-user@shakespeare.lit/cool-client"
                to="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscriptions node="dev" />
              </pubsub>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                to="emacs-user@shakespeare.lit/cool-client"
                from="aim.shakespeare.lit"
                id="0">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <subscriptions node="dev">
                  <subscription node='dev'
                                jid='emacs-user@shakespeare.lit'
                                subscription='subscribed' />
                </subscriptions>
              </pubsub>
            </iq>
            """
        )

    def test_get_updated_spaces(self) -> None:
        spaces = self.run_coro(_get_updated_spaces(self.user_session()))
        assert len(spaces)
        assert [s.name for s in spaces] == [
            "Sales team",
            "The bosses lounge",
            "The developers corner",
        ]

    def test_fetch_space_metadata_onlyonce(self) -> None:
        req = """
            <iq type='get'
                from='emacs-user@shakespeare.lit/cool-client'
                to='aim.shakespeare.lit'
                id='{id}'>
              <query xmlns='http://jabber.org/protocol/disco#info'
                     node='dev' />
            </iq>
            """
        with unittest.mock.patch(
            "test_spaces.Bookmarks.fetch_space_metadata", return_value=SpaceMetadata()
        ) as fetch_space_metadata:
            self.recv(req.format(id=1))
            self.recv(req.format(id=2))
        fetch_space_metadata.assert_awaited_once()



import contextlib
import logging
import typing

from slixmpp import JID, Presence
from slixmpp.exceptions import XMPPError

from ...contact import LegacyContact
from ...contact.roster import ContactIsUser
from ...util.types import AnySession
from ...util.util import merge_resources
from .util import DispatcherMixin, exceptions_to_xmpp_errors

if typing.TYPE_CHECKING:
    from slidge.util.types import AnyGateway


class _IsDirectedAtComponent(Exception):
    def __init__(self, session: AnySession) -> None:
        self.session = session


class PresenceHandlerMixin(DispatcherMixin):
    __slots__: list[str] = []
    xmpp: "AnyGateway"

    def __init__(self, xmpp: "AnyGateway") -> None:
        super().__init__(xmpp)

        xmpp.add_event_handler("presence_subscribe", self._handle_subscribe)
        xmpp.add_event_handler("presence_subscribed", self._handle_subscribed)
        xmpp.add_event_handler("presence_unsubscribe", self._handle_unsubscribe)
        xmpp.add_event_handler("presence_unsubscribed", self._handle_unsubscribed)
        xmpp.add_event_handler("presence_probe", self._handle_probe)
        xmpp.add_event_handler("presence", self.on_presence)

    async def __get_contact(self, pres: Presence) -> LegacyContact:
        sess = await self._get_session(pres)
        pto = pres.get_to()
        if pto == self.xmpp.boundjid.bare:
            raise _IsDirectedAtComponent(sess)
        await sess.contacts.ready
        return await sess.contacts.by_jid(pto)  # type:ignore[no-any-return]

    @exceptions_to_xmpp_errors
    async def _handle_subscribe(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            pres.reply().send()
            return

        if contact.is_friend:
            pres.reply().send()
        else:
            await contact.on_friend_request(pres["status"])

    @exceptions_to_xmpp_errors
    async def _handle_unsubscribe(self, pres: Presence) -> None:
        pres.reply().send()

        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent as e:
            e.session.send_gateway_message("Bye bye!")
            await e.session.kill_by_jid(e.session.user_jid)
            return

        contact.is_friend = False
        await contact.on_friend_delete(pres["status"])

    @exceptions_to_xmpp_errors
    async def _handle_subscribed(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            return

        await contact.on_friend_accept()
        contact.send_last_presence(force=True)

    @exceptions_to_xmpp_errors
    async def _handle_unsubscribed(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            return

        if contact.is_friend:
            contact.is_friend = False
            await contact.on_friend_delete(pres["status"])

    @exceptions_to_xmpp_errors
    async def _handle_probe(self, pres: Presence) -> None:
        try:
            contact = await self.__get_contact(pres)
        except _IsDirectedAtComponent:
            session = await self._get_session(pres)
            session.send_cached_presence(pres.get_from())
            return
        if contact.is_friend:
            contact.send_last_presence(force=True)
        else:
            reply = pres.reply()
            reply["type"] = "unsubscribed"
            reply.send()

    @exceptions_to_xmpp_errors
    async def on_presence(self, p: Presence) -> None:
        if "muc_join" in p:
            # handled in on_groupchat_join
            # without this early return, since we switch from and to in this
            # presence stanza, on_groupchat_join ends up trying to instantiate
            # a MUC with the user's JID, which in turn leads to slidge sending
            # a (error) presence from=the user's JID, which terminates the
            # XML stream.
            return

        session = await self._get_session(p)

        pto = p.get_to()
        if pto == self.xmpp.boundjid.bare:
            await self._on_presence_to_component(session, p)
            return

        if p.get_type() == "available":
            try:
                contact = await session.contacts.by_jid(pto)
            except XMPPError:
                contact = None
            except ContactIsUser:
                raise XMPPError(
                    "bad-request", "Actions with yourself are not supported."
                )
            if contact is not None:
                await self.xmpp.pubsub.on_presence_available(p, contact)
                return

        muc = session.bookmarks.by_jid_only_if_exists(JID(pto.bare))

        if muc is not None and p.get_type() == "unavailable":
            muc.on_presence_unavailable(p)
            return

        if muc is None or p.get_from().resource not in muc.get_user_resources():
            return

        if pto.resource == muc.user_nick:
            # Ignore presence stanzas with the valid nick.
            # even if joined to the group, we might receive those from clients,
            # when setting a status message, or going away, etc.
            return

        # We can't use XMPPError here because XMPPError does not have a way to
        # add the <x xmlns="http://jabber.org/protocol/muc" /> element

        error_stanza = p.error()
        error_stanza.set_to(p.get_from())
        error_stanza.set_from(pto)
        error_stanza.enable("muc_join")  # <x xmlns="http://jabber.org/protocol/muc" />
        error_stanza.enable("error")
        error_stanza["error"]["type"] = "cancel"
        error_stanza["error"]["by"] = muc.jid
        error_stanza["error"]["condition"] = "not-acceptable"
        error_stanza["error"]["text"] = (
            "Slidge does not let you change your nickname in groups."
        )
        error_stanza.send()

    async def _on_presence_to_component(self, session: AnySession, p: Presence) -> None:
        session.log.debug("Received a presence from %s", p.get_from())
        if (ptype := p.get_type()) not in _USEFUL_PRESENCES:
            return
        if not session.user.preferences.get("sync_presence", False):
            session.log.debug("User does not want to sync their presence")
            return
        # NB: get_type() returns either a proper presence type or
        #     a presence show if available. Weird, weird, weird slix.
        resources = self.xmpp.roster[self.xmpp.boundjid.bare][p.get_from()].resources
        with contextlib.suppress(NotImplementedError):
            await session.on_presence(
                p.get_from().resource,
                ptype,  # type: ignore
                p["status"],
                resources,
                merge_resources(resources),
            )
        if p.get_type() == "available":
            await self.xmpp.pubsub.on_presence_available(p, None)
            for contact in session.contacts:
                await self.xmpp.pubsub.on_presence_available(p, contact)


_USEFUL_PRESENCES = {"available", "unavailable", "away", "chat", "dnd", "xa"}

log = logging.getLogger(__name__)



from slixmpp.exceptions import XMPPError

from slidge import (
    BaseGateway,
    BaseSession,
    LegacyBookmarks,
    LegacyContact,
    LegacyParticipant,
    LegacyRoster,
)
from slidge.util.test import SlidgeTest
from slidge.util.types import MessageReference


class Session(BaseSession):
    async def login(self):
        pass


class Gateway(BaseGateway):
    COMPONENT_NAME = "Megacorp Chat XMPP Gateway"
    GROUPS = True


class Contact(LegacyContact):
    pass


class Roster(LegacyRoster):
    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        if not legacy_id.startswith("room-"):
            raise XMPPError
        return legacy_id


class Bookmarks(LegacyBookmarks):
    async def fill(self):
        self.user_nick = "🎉coquinou"


class TestMamCorrection(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        self._add_slidge_user()

    def test_reply_to_user(self) -> None:
        muc = self.get_joined_muc("room-coquinou")
        part: LegacyParticipant = self.run_coro(muc.get_user_participant())
        part.send_text(
            "a body",
            "a-legacy-id",
            reply_to=MessageReference(
                "ref",
                author="user",
                body="blabla",
            ),
        )
        self.send(  # language=XML
            """
            <message type="groupchat"
                     from="room-coquinou@aim.shakespeare.lit/coquinou-1u25g"
                     id="a-legacy-id"
                     to="romeo@shakespeare.lit/gajim">
              <body>&gt; 🎉coquinou:\n&gt; blabla\na body</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="a-legacy-id"
                         by="room-coquinou@aim.shakespeare.lit" />
              <reply xmlns="urn:xmpp:reply:0"
                     id="ref"
                     to="room-coquinou@aim.shakespeare.lit/coquinou-1u25g" />
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:reply:0">
                <body start="0"
                      end="22" />
              </fallback>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
              <nick xmlns="http://jabber.org/protocol/nick">🎉coquinou</nick>
            </message>
            """
        )



import logging
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
import sqlalchemy.orm
from slixmpp import JID
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004.stanza import Form  # type:ignore[attr-defined]
from slixmpp.plugins.xep_0030.stanza.info import DiscoInfo
from slixmpp.plugins.xep_0030.stanza.items import DiscoItems
from slixmpp.types import OptJid

from ...db.models import Room, Space
from ...util.types import AnySession
from .util import DispatcherMixin

if TYPE_CHECKING:
    from slidge.util.types import AnyGateway, AnySession


class DiscoMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: "AnyGateway") -> None:
        super().__init__(xmpp)

        xmpp.plugin["xep_0030"].set_node_handler(
            "get_info",
            jid=None,
            node=None,
            handler=self.get_info,
        )

        xmpp.plugin["xep_0030"].set_node_handler(
            "get_items",
            jid=None,
            node=None,
            handler=self.get_items,
        )

    async def get_info(
        self,
        jid: OptJid,
        node: str | None,
        ifrom: OptJid,
        data: Any,  # noqa:ANN401
    ) -> DiscoInfo | None:
        # TODO: OMG the control flow in this method 🤮
        if ifrom == self.xmpp.boundjid.bare or (
            jid in (self.xmpp.boundjid.bare, None)
            and (not node or node.startswith("http://slixmpp.com/ver/"))
        ):
            return self.xmpp.plugin["xep_0030"].static.get_info(jid, node, ifrom, data)

        if ifrom is None:
            raise XMPPError("subscription-required")

        assert jid is not None
        session = await self._get_session_from_jid(jid=ifrom)

        if not jid.username:
            assert node is not None  # 🤮
            return await self.__spaces_info(session, node)

        log.debug("Looking for entity: %s", jid)

        entity = await session.get_contact_or_group_or_participant(jid)

        if entity is None:
            raise XMPPError("item-not-found")

        return await entity.get_disco_info(jid, node)

    async def __spaces_info(self, session: AnySession, node: str) -> DiscoInfo:
        legacy_id = await session.bookmarks.space_node_to_legacy_id(node)
        with self.xmpp.store.session(expire_on_commit=False) as orm:
            space = self.xmpp.store.spaces.get_by_legacy_id(
                orm, session.user_pk, legacy_id, full=True
            )
            if space is None:
                raise XMPPError("item-not-found", f"No space for node '{node}'")
        space = await session.bookmarks.update_space_if_needed(space)
        form = Form()
        form["type"] = "result"
        form.add_field(
            var="FORM_TYPE",
            ftype="hidden",
            value="http://jabber.org/protocol/pubsub#meta-data",
        )
        form.add_field(
            var="pubsub#type",
            value="urn:xmpp:spaces:0",
        )
        form.add_field(
            var="pubsub#creator",
            type="jid-single",
            value=self.xmpp.boundjid.bare
            if space.creator is None
            else str(space.creator.jid),
        )
        form.add_field(var="pubsub#title", value=space.name)
        form.add_field(
            var="pubsub#owner",
            type="jid-multi",
            value=[str(owner.jid) for owner in space.owners],
        )
        if space.member_count is not None:
            form.add_field(var="pubsub#num_subscribers", value=str(space.member_count))
        if space.description is not None:
            form.add_field(var="pubsub#description", value=space.description)
        info = DiscoInfo()
        info.add_identity(category="pubsub", itype="leaf")
        info.add_feature("http://jabber.org/protocol/pubsub")
        for feat in (
            "meta-data",
            "item-ids",
            "manage-subscriptions",
            "modify-affiliations",
            "outcast-affiliation",
            "retract-items",
            "retrieve-affiliations",
            "retrieve-items",
            "retrieve-subscriptions",
            "subscribe",
            "subscription-notifications",
        ):
            info.add_feature(f"http://jabber.org/protocol/pubsub#{feat}")
        info.append(form)
        return info

    async def get_items(
        self,
        jid: OptJid,
        node: str | None,
        ifrom: OptJid,
        data: Any,  # noqa:ANN401
    ) -> DiscoItems:
        if ifrom is None:
            raise XMPPError("bad-request")

        assert ifrom is not None
        session = await self._get_session_from_jid(ifrom)

        if jid == self.xmpp.boundjid.bare or not jid:
            if node:
                return DiscoItems()
            else:
                return await self.__list_rooms(session, data)

        entity = await session.get_contact_or_group_or_participant(jid)

        if entity is None:
            raise XMPPError("item-not-found")

        return await entity.get_disco_items(node)

    async def __list_rooms(
        self,
        session: "AnySession",
        data: dict,  # type:ignore[type-arg]
    ) -> DiscoItems:
        try:
            included_types = data["disco_items"]["filter"]["included_types"]
        except KeyError:
            included_types = []

        d = DiscoItems()
        with self.xmpp.store.session() as orm:
            if included_types == ["urn:xmpp:spaces:0"]:
                if not self.xmpp.SPACES:
                    raise XMPPError(
                        "feature-not-implemented",
                        "This gateway does not support XEP-0503",
                    )
                await self.__spaces(orm, session, d)
            elif not included_types:
                self.__rooms(orm, session, d)
            else:
                raise XMPPError("item-not-found", f"No items for {included_types}")

        return d

    async def __spaces(
        self, orm: sqlalchemy.orm.Session, session: AnySession, d: DiscoItems
    ) -> None:
        await session.bookmarks.update_spaces_info()
        for space in orm.execute(
            sa.select(Space)
            .options(sa.orm.load_only(Space.legacy_id, Space.name))
            .filter_by(user=session.user)
        ).scalars():
            d.add_item(
                JID(self.xmpp.boundjid.bare),
                name=space.name,
                node=await session.bookmarks.space_legacy_id_to_node(space.legacy_id),
            )

    @staticmethod
    def __rooms(
        orm: sqlalchemy.orm.Session, session: AnySession, d: DiscoItems
    ) -> None:
        for room in orm.execute(
            sa.select(Room)
            .options(sa.orm.load_only(Room.jid, Room.name))
            .filter_by(user=session.user)
            .order_by(Room.name)
        ).scalars():
            d.add_item(room.jid, name=room.name)


log = logging.getLogger(__name__)



import abc
import logging
from collections.abc import Iterator
from typing import Any, Generic, Literal, overload

from slixmpp import JID
from slixmpp.exceptions import XMPPError
from sqlalchemy.orm import Session as OrmSession

from slidge.contact import LegacyContact

from ..db.models import Contact, Room, Space
from ..util import SubclassableOnce
from ..util.jid_escaping import EscapeMixin
from ..util.lock import NamedLockMixin
from ..util.types import (
    AnyMUC,
    AnySession,
    LegacyMUCType,
    SpaceMetadata,
)


class LegacyBookmarks(
    Generic[LegacyMUCType],
    EscapeMixin,
    NamedLockMixin,
    SubclassableOnce,
):
    """
    This is instantiated once per :class:`~slidge.BaseSession`
    """

    _muc_cls: type[LegacyMUCType]

    def __init__(self, session: AnySession) -> None:
        self.session = session
        self.xmpp = session.xmpp
        self.user_jid = session.user_jid

        self._user_nick: str = self.session.user_jid.node

        super().__init__()
        self.log = logging.getLogger(f"{self.user_jid.bare}:bookmarks")
        self.ready = self.session.xmpp.loop.create_future()
        if not self.xmpp.GROUPS:
            self.ready.set_result(True)

    @property
    def user_nick(self) -> str:
        return self._user_nick

    @user_nick.setter
    def user_nick(self, nick: str) -> None:
        self._user_nick = nick

    def orm(
        self,
        **kwargs: Any,  # noqa:ANN401
    ) -> OrmSession:
        return self.session.xmpp.store.session(**kwargs)

    def from_store(self, stored: Room) -> LegacyMUCType:
        return self._muc_cls(self.session, stored)

    def __iter__(self) -> Iterator[LegacyMUCType]:
        with self.xmpp.store.session() as orm:
            for stored in orm.query(Room).filter_by(user=self.session.user).all():
                if stored.updated:
                    yield self.from_store(stored)

    def __repr__(self) -> str:
        return f"<Bookmarks of {self.user_jid}>"

    async def legacy_id_to_jid_local_part(self, legacy_id: str) -> str:
        return await self.legacy_id_to_jid_username(legacy_id)

    async def jid_local_part_to_legacy_id(self, local_part: str) -> str:
        return await self.jid_username_to_legacy_id(local_part)

    async def by_jid(self, jid: JID) -> LegacyMUCType:
        if jid.resource:
            jid = JID(jid.bare)
        async with self.lock(("bare", jid.bare)):
            legacy_id = await self.jid_local_part_to_legacy_id(jid.node)
            if self.get_lock(("legacy_id", legacy_id)):
                self.session.log.debug("Already updating %s via by_legacy_id()", jid)
                return await self.by_legacy_id(legacy_id)

            with self.session.xmpp.store.session() as orm:
                stored = (
                    orm.query(Room)
                    .filter_by(user_account_id=self.session.user_pk, jid=jid)
                    .one_or_none()
                )
            if stored is None:
                stored = Room(
                    user_account_id=self.session.user_pk,
                    jid=jid,
                    legacy_id=legacy_id,
                )
            return await self.__update_if_needed(stored)

    def by_jid_only_if_exists(self, jid: JID) -> LegacyMUCType | None:
        with self.xmpp.store.session(expire_on_commit=False) as orm:
            stored = (
                orm.query(Room).filter_by(user=self.session.user, jid=jid).one_or_none()
            )
            if stored is not None and stored.updated:
                return self.from_store(stored)
        return None

    @overload
    async def by_legacy_id(self, legacy_id: str) -> "LegacyMUCType": ...

    @overload
    async def by_legacy_id(
        self, legacy_id: str, create: Literal[False]
    ) -> "LegacyMUCType | None": ...

    @overload
    async def by_legacy_id(
        self, legacy_id: str, create: Literal[True]
    ) -> "LegacyMUCType": ...

    async def by_legacy_id(
        self, legacy_id: str, create: bool = False
    ) -> LegacyMUCType | None:
        async with self.lock(("legacy_id", legacy_id)):
            local = await self.legacy_id_to_jid_local_part(legacy_id)
            jid = JID(f"{local}@{self.xmpp.boundjid}")
            if self.get_lock(("bare", jid.bare)):
                self.session.log.debug("Already updating %s via by_jid()", jid)
                if create:
                    return await self.by_jid(jid)
                else:
                    return self.by_jid_only_if_exists(jid)

            with self.xmpp.store.session() as orm:
                stored = (
                    orm.query(Room)
                    .filter_by(
                        user_account_id=self.session.user_pk,
                        legacy_id=str(legacy_id),
                    )
                    .one_or_none()
                )
            if stored is None:
                stored = Room(
                    user_account_id=self.session.user_pk,
                    jid=jid,
                    legacy_id=str(legacy_id),
                )
            return await self.__update_if_needed(stored)

    async def __update_if_needed(self, stored: Room) -> LegacyMUCType:
        muc = self.from_store(stored)
        if muc.stored.updated:
            return muc

        with muc.updating_info():
            try:
                await muc.update_info()
            except NotImplementedError:
                pass
            except XMPPError:
                raise
            except Exception as e:
                raise XMPPError("internal-server-error", str(e))
        muc.archive.room = muc.stored
        if self.ready.done() and muc.stored.space_id:
            with self.orm() as orm:
                orm.add(muc.stored)
                self.xmpp.pubsub.broadcast_space(self.session, muc.stored.space)
        return muc

    async def fill(self) -> None:
        """
        Establish a user's known groups.

        This has to be overridden in plugins with group support and at the
        minimum, this should ``await self.by_legacy_id(group_id)`` for all
        the groups a user is part of.

        Slidge internals will call this on successful :meth:`BaseSession.login`

        """
        if self.xmpp.GROUPS:
            raise NotImplementedError(
                "The plugin advertised support for groups but"
                " LegacyBookmarks.fill() was not overridden."
            )

    async def remove(
        self,
        muc: AnyMUC,
        reason: str = "You left this group from the official client.",
        kick: bool = True,
    ) -> None:
        """
        Delete everything about a specific group.

        This should be called when the user leaves the group from the official
        app.

        :param muc: The MUC to remove.
        :param reason: Optionally, a reason why this group was removed.
        :param kick: Whether the user should be kicked from this group. Set this
            to False in case you do this somewhere else in your code, eg, on
            receiving the confirmation that the group was deleted.
        """
        if kick:
            user_participant = await muc.get_user_participant()
            user_participant.kick(reason)
        with self.xmpp.store.session() as orm:
            orm.add(muc.stored)
            orm.refresh(muc.stored)
            orm.delete(muc.stored)
            orm.commit()

    async def update_space_if_needed(self, space: Space) -> Space:
        async with self.lock(("space", space.legacy_id)):
            with self.orm() as orm:
                orm.add(space)
                orm.refresh(space)
                if space.updated:
                    return space
            meta: SpaceMetadata = await self.fetch_space_metadata(space.legacy_id)
        return await self.__update_space_metadata(space, meta)

    async def __update_space_metadata(self, space: Space, meta: SpaceMetadata) -> Space:
        creator = (
            await self.__get_stored_contact(meta.creator_legacy_id)
            if meta.creator_legacy_id
            else None
        )
        owners: list[Contact] = []
        for legacy_id in set(meta.owner_legacy_ids):
            if legacy_id == meta.creator_legacy_id:
                # We don't want to fetch any contact twice here to avoid:
                # Can't attach instance <Contact at xxx>; another instance [...] is already present in this session.
                continue
            owner = await self.__get_stored_contact(legacy_id)
            if owner is not None:
                owners.append(owner)
        if (
            creator is not None
            and meta.creator_legacy_id
            and meta.creator_legacy_id in meta.owner_legacy_ids
        ):
            owners.append(creator)
        with self.orm(expire_on_commit=False) as orm:
            if creator is not None:
                creator = orm.merge(creator)
            owners = [orm.merge(owner) for owner in owners]
            space = orm.merge(space)
            space.name = meta.name or space.legacy_id
            space.creator = creator
            space.owners = owners
            space.description = meta.description
            space.member_count = meta.member_count
            space.updated = True
            orm.commit()
        return space

    async def __get_stored_contact(self, legacy_id_str: str) -> Contact | None:
        try:
            contact: LegacyContact = await self.session.contacts.by_legacy_id(
                legacy_id_str
            )
        except Exception as e:
            self.log.warning("Could not get contact: %r", e)
            return None
        return contact.stored

    @abc.abstractmethod
    async def fetch_space_metadata(self, legacy_id: str) -> SpaceMetadata:
        """
        Fetch metadata associated to a space.

        This is called once per slidge runtime. It should return metadata
        associated to the space identified by its ``legacy_id``.
        If there are updates to this metata, they should be communicated to
        slidge by calling :func:`LegacyBookmarks.update_space_metadata`.

        :param legacy_id: Identifier of the space.

        :return: Metadata associated to the space.
        """
        raise NotImplementedError

    async def update_spaces_info(self) -> None:
        with self.orm() as orm:
            spaces = self.session.xmpp.store.spaces.get_unupdated(
                orm, self.session.user_pk
            )
        for space in spaces:
            await self.update_space_if_needed(space)

    async def space_legacy_id_to_node(self, legacy_id: str) -> str:
        return str(legacy_id)

    async def space_node_to_legacy_id(self, node: str) -> str:
        return node

    async def update_space_metadata(
        self,
        legacy_id: str,
        metadata: SpaceMetadata,
    ) -> None:
        """
        Updates metadata associated to a space.

        :param legacy_id: Identifier of the space.
        :param name: Metadata associated to this space.
        """
        with self.orm(expire_on_commit=False) as orm:
            space = self.session.xmpp.store.spaces.add_or_get(
                orm,
                self.session.user_pk,
                str(legacy_id),
            )
        space = await self.__update_space_metadata(space, metadata)
        self.xmpp.pubsub.broadcast_space_metadata(
            self.session, space, await self.space_legacy_id_to_node(legacy_id)
        )



import asyncio
from collections.abc import Awaitable
from pathlib import Path
from typing import Any

ASSETS_DIR = Path(__file__).parent / "assets"
if not ASSETS_DIR.exists():
    ASSETS_DIR = Path(__file__).parent.parent / "dev" / "assets"
    if not ASSETS_DIR.exists():
        raise FileNotFoundError


def later(awaitable: Awaitable[Any]) -> None:
    asyncio.create_task(_later(awaitable))


async def _later(awaitable: Awaitable[Any]) -> None:
    await asyncio.sleep(0.5)
    await awaitable



"""
User actions
"""

import typing

from slidge.core.session import BaseSession
from slidge.db.models import GatewayUser

from .legacy_client import SuperDuperClient

if typing.TYPE_CHECKING:
    from .contact import Contact  # noqa:F401


class Session(BaseSession["Contact"]):
    def __init__(self, user: GatewayUser) -> None:
        super().__init__(user)
        self.legacy_client = SuperDuperClient(self)
        self.contacts.user_legacy_id = str(self.legacy_client.user_id)

    async def login(self) -> str:
        await self.legacy_client.login()
        return "Success!"



import logging
import uuid
import warnings
from collections.abc import Collection, Iterator
from copy import copy
from datetime import UTC, datetime

from slixmpp import Iq, Message

from ..db.models import ArchivedMessage, ArchivedMessageSource, Room
from ..db.store import SlidgeStore
from ..util.archive_msg import HistoryMessage
from ..util.types import AnyParticipant, HoleBound


class MessageArchive:
    def __init__(self, room: Room, store: SlidgeStore) -> None:
        self.room = room
        self.__rooms_store = store.rooms
        self.__store = store.mam

    def add(
        self,
        msg: Message,
        participant: "AnyParticipant | None" = None,
        archive_only: bool = False,
        legacy_msg_id: str | None = None,
    ) -> None:
        """
        Add a message to the archive if it is deemed archivable

        :param msg:
        :param participant:
        :param archive_only:
        :param legacy_msg_id:
        """
        if not archivable(msg):
            return
        new_msg = copy(msg)
        if participant and not participant.muc.is_anonymous:
            new_msg["muc"]["role"] = participant.role or "participant"
            new_msg["muc"]["affiliation"] = participant.affiliation or "member"
            if participant.contact:
                new_msg["muc"]["jid"] = participant.contact.jid.bare
            elif participant.is_user:
                new_msg["muc"]["jid"] = participant.user_jid.bare
            elif participant.is_system:
                new_msg["muc"]["jid"] = participant.muc.jid
            else:
                warnings.warn(
                    f"No real JID for participant '{participant.nickname}' in '{self.room.name}'"
                )
                new_msg["muc"]["jid"] = (
                    f"{uuid.uuid4()}@{participant.xmpp.boundjid.bare}"
                )

        if "replace" in msg:
            legacy_msg_id = None

        with self.__store.session(expire_on_commit=False) as orm:
            if self.room.id is None:
                self.room = self.__rooms_store.get(
                    orm, self.room.user_account_id, self.room.legacy_id
                )
            self.__store.add_message(
                orm,
                self.room.id,
                HistoryMessage(new_msg),
                archive_only,
                None if legacy_msg_id is None else str(legacy_msg_id),
            )
            orm.commit()

    def __iter__(self) -> Iterator[HistoryMessage]:
        return iter(self.get_all())

    @staticmethod
    def __to_bound(stored: ArchivedMessage) -> HoleBound:
        return HoleBound(
            stored.legacy_id,  # type:ignore
            stored.timestamp.replace(tzinfo=UTC),
        )

    def get_hole_bounds(self) -> tuple[HoleBound | None, HoleBound | None]:
        with self.__store.session() as orm:
            most_recent = self.__store.get_most_recent_with_legacy_id(orm, self.room.id)
            if most_recent is None:
                return None, None
            if most_recent.source == ArchivedMessageSource.BACKFILL:
                # most recent = only backfill, fetch everything since last backfill
                return self.__to_bound(most_recent), None

            most_recent_back_filled = self.__store.get_most_recent_with_legacy_id(
                orm, self.room.id, ArchivedMessageSource.BACKFILL
            )
            if most_recent_back_filled is None:
                # group was never back-filled, fetch everything before first live
                least_recent_live = self.__store.get_first(orm, self.room.id, True)
                assert least_recent_live is not None
                return None, self.__to_bound(least_recent_live)

            assert most_recent_back_filled.legacy_id is not None
            least_recent_live = self.__store.get_least_recent_with_legacy_id_after(
                orm, self.room.id, most_recent_back_filled.legacy_id
            )
            assert least_recent_live is not None
            # this is a hole caused by slidge downtime
            return self.__to_bound(most_recent_back_filled), self.__to_bound(
                least_recent_live
            )

    def get_all(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        before_id: str | None = None,
        after_id: str | None = None,
        ids: Collection[str] = (),
        last_page_n: int | None = None,
        sender: str | None = None,
        flip: bool = False,
    ) -> Iterator[HistoryMessage]:
        with self.__store.session() as orm:
            yield from self.__store.get_messages(
                orm,
                self.room.id,
                before_id=before_id,
                after_id=after_id,
                ids=ids,
                last_page_n=last_page_n,
                sender=sender,
                start_date=start_date,
                end_date=end_date,
                flip=flip,
            )

    async def send_metadata(self, iq: Iq) -> None:
        """
        Send archive extent, as per the spec

        :param iq:
        :return:
        """
        reply = iq.reply()
        with self.__store.session() as orm:
            messages = self.__store.get_first_and_last(orm, self.room.id)
        if messages:
            for x, m in [("start", messages[0]), ("end", messages[-1])]:
                reply["mam_metadata"][x]["id"] = m.id
                reply["mam_metadata"][x]["timestamp"] = m.sent_on.replace(tzinfo=UTC)
        else:
            reply.enable("mam_metadata")
        reply.send()


def archivable(msg: Message) -> bool:
    """
    Determine if a message stanza is worth archiving, ie, convey meaningful
    info

    :param msg:
    :return:
    """

    if "no-store" in msg:
        return False

    if "no-permanent-store" in msg:
        return False

    if "store" in msg:
        return True

    if msg["body"]:
        return True

    if "retract" in msg:
        return True

    if "reactions" in msg:
        return True

    if "displayed" in msg:
        return True

    return bool(msg["thread"] and msg["subject"])


log = logging.getLogger(__name__)



import base64
import io
import logging
import uuid
import warnings
from collections.abc import Iterable
from datetime import UTC, datetime
from pathlib import Path
from typing import cast

from PIL import Image
from slixmpp import Message
from slixmpp.plugins.xep_0511.stanza import LinkMetadata
from slixmpp.types import MessageTypes

from slidge.util import strip_illegal_chars

from ...db.models import GatewayUser
from ...util.types import (
    AnyMUC,
    AnyParticipant,
    ChatState,
    LinkPreview,
    MessageReference,
    ProcessingHint,
)
from .. import config
from .base import BaseSender


class MessageMaker(BaseSender):
    mtype: MessageTypes = NotImplemented
    _can_send_carbon: bool = NotImplemented
    STRIP_SHORT_DELAY = False
    USE_STANZA_ID = False

    muc: AnyMUC

    def _recipient_pk(self) -> int:
        return (
            self.muc.stored.id if self.is_participant else self.stored.id  # type:ignore
        )

    def _make_message(
        self,
        state: ChatState | None = None,
        hints: Iterable[ProcessingHint] = (),
        legacy_msg_id: str | None = None,
        when: datetime | None = None,
        reply_to: MessageReference | None = None,
        carbon: bool = False,
        link_previews: Iterable[LinkPreview] | None = None,
        **kwargs: object,
    ) -> Message:
        body = kwargs.pop("mbody", None)
        mfrom = kwargs.pop("mfrom", self.jid)
        mto = kwargs.pop("mto", None)
        thread = kwargs.pop("thread", None)
        # the msg needs to have jabber:client as xmlns, so
        # we don't want to associate with the XML stream
        msg_cls = Message if carbon and self._can_send_carbon else self.xmpp.Message
        msg = msg_cls(
            sfrom=mfrom,
            stype=kwargs.pop("mtype", None) or self.mtype,
            sto=mto,
            **kwargs,
        )
        if body:
            assert isinstance(body, str)
            msg["body"] = strip_illegal_chars(body, "�")
            state = "active"
        if thread:
            with self.xmpp.store.session() as orm:
                thread_str = str(thread)
                msg["thread"] = (
                    self.xmpp.store.id_map.get_thread(
                        orm,
                        self._recipient_pk(),
                        thread_str,
                        self.is_participant,
                    )
                    or thread_str
                )
        if state:
            msg["chat_state"] = state
        for hint in hints:
            msg.enable(hint)
        self._set_msg_id(msg, legacy_msg_id)
        self._add_delay(msg, when)
        if link_previews:
            self._add_link_previews(msg, link_previews)
        if reply_to:
            self._add_reply_to(msg, reply_to)
        return msg

    def _set_msg_id(self, msg: Message, legacy_msg_id: str | None = None) -> None:
        if legacy_msg_id is not None:
            msg.set_id(legacy_msg_id)
            if self.USE_STANZA_ID:
                msg["stanza_id"]["id"] = legacy_msg_id
                msg["stanza_id"]["by"] = self.muc.jid
        elif self.USE_STANZA_ID:
            msg["stanza_id"]["id"] = str(uuid.uuid4())
            msg["stanza_id"]["by"] = self.muc.jid

    def _legacy_to_xmpp(self, legacy_id: str) -> list[str]:
        with self.xmpp.store.session() as orm:
            ids = self.xmpp.store.id_map.get_xmpp(
                orm,
                self._recipient_pk(),
                str(legacy_id),
                self.is_participant,
            )
            if ids:
                return ids
        return [legacy_id]

    def _add_delay(self, msg: Message, when: datetime | None) -> None:
        if when:
            if when.tzinfo is None:
                when = when.astimezone(UTC)
            if self.STRIP_SHORT_DELAY:
                delay = (datetime.now().astimezone(UTC) - when).seconds
                if delay < config.IGNORE_DELAY_THRESHOLD:
                    return
            msg["delay"].set_stamp(when)
            msg["delay"].set_from(self.xmpp.boundjid.bare)

    def _add_reply_to(self, msg: Message, reply_to: MessageReference) -> None:
        xmpp_id = self._legacy_to_xmpp(reply_to.legacy_id)[0]
        msg["reply"]["id"] = xmpp_id

        muc = getattr(self, "muc", None)

        if entity := reply_to.author:
            if entity == "user" or isinstance(entity, GatewayUser):
                if isinstance(entity, GatewayUser):
                    warnings.warn(
                        "Using a GatewayUser as the author of a "
                        "MessageReference is deprecated. Use the string 'user' "
                        "instead.",
                        DeprecationWarning,
                    )
                if muc:
                    msg["reply"]["to"] = muc.user_muc_jid
                    fallback_nick = muc.user_nick
                else:
                    msg["reply"]["to"] = self.session.user_jid
                    # TODO: here we should use preferably use the PEP nick of the user
                    # (but it doesn't matter much)
                    fallback_nick = self.session.user_jid.user
            else:
                if muc:
                    if hasattr(entity, "muc"):
                        # TODO: accept a Contact here and use muc.get_participant_by_legacy_id()
                        # a bit of work because right now this is a sync function
                        entity = cast(AnyParticipant, entity)
                        fallback_nick = entity.nickname
                    else:
                        warnings.warn(
                            "The author of a message reference in a MUC must be a"
                            " Participant instance, not a Contact"
                        )
                        fallback_nick = entity.name
                else:
                    fallback_nick = entity.name
                msg["reply"]["to"] = entity.jid
        else:
            fallback_nick = None

        if fallback := reply_to.body:
            msg["reply"].add_quoted_fallback(fallback, fallback_nick)

    def _add_link_previews(
        self, msg: Message, link_previews: Iterable[LinkPreview]
    ) -> None:
        for preview in link_previews:
            if preview.is_empty:
                continue
            element = LinkMetadata()
            for i, name in enumerate(preview._fields):
                val = preview[i]
                if isinstance(val, Path):
                    val = val.read_bytes()
                if isinstance(val, bytes):
                    val = self._process_link_preview_image(val)
                if not val:
                    continue
                element[name] = val
            msg.append(element)

    @staticmethod
    def _process_link_preview_image(data: bytes) -> str | None:
        # this will block the main thread. if this proves to be an issue in practice,
        # this could be rewritten to use the thread pool we use to resize avatars.
        try:
            image = Image.open(io.BytesIO(data))
        except Exception:
            log.exception("Skipping link preview image")
            return None

        rewrite = False
        if image.format != "JPEG":
            rewrite = True

        if any(x > MAX_LINK_PREVIEW_IMAGE_SIZE for x in image.size):
            image.thumbnail((MAX_LINK_PREVIEW_IMAGE_SIZE, MAX_LINK_PREVIEW_IMAGE_SIZE))
            rewrite = True

        if rewrite:
            with io.BytesIO() as f:
                image.save(f, format="JPEG")
                data = f.getvalue()

        return "data:image/jpeg;base64," + base64.b64encode(data).decode("utf-8")


# Instead of having a hardcoded value for this, we would ideally use
# XEP-0478: Stream Limits Advertisement to know which size is authorized.
# However, this isn't possible until XEP-0225: Component Connections is a thing.
# Prosody defaults to 512kb for s2s connection and 10Mb for c2s connections.
# Some quick tests about JPEG image:
# median size of 50 base64-encoded JPEG random RGB image
# 128x128 pixels: 14kb ± 0.04
# 256x256 pixels: 53kb ± 0.06  # sounds like a good tradeoff
# 384x384 pixels: 119kb ± 0.11
# 512x512 pixels: 211kb ± 0.13
# 640x640 pixels: 329kb ± 0.15
# 768x768 pixels: 473kb ± 0.21
# 896x896 pixels: 644kb ± 0.27
# 1024x1024 pixels: 841kb ± 0.22
MAX_LINK_PREVIEW_IMAGE_SIZE = 256

log = logging.getLogger(__name__)



import logging
from collections.abc import Iterable
from typing import Any, Generic, TypeVar

from slixmpp import JID
from slixmpp.types import JidStr, OptJidStr, RosterState

GatewayType = TypeVar("GatewayType")


class YesSet(set):
    """
    A pseudo-set which always test True for membership
    """

    def __contains__(self, item: object) -> bool:
        log.debug("Test in")
        return True


class RosterBackend(Generic[GatewayType]):
    """
    A pseudo-roster for the gateway component.

    If a user is in the user store, this will behave as if the user is part of the
    roster with subscription "both", and "none" otherwise.

    This is rudimentary but the only sane way I could come up with so far.
    """

    xmpp: GatewayType

    def __init__(self, xmpp: GatewayType) -> None:
        self.xmpp = xmpp

    def entries(
        self, owner: OptJidStr, db_state: dict[str, Any] | None = None
    ) -> Iterable[str]:
        return YesSet()

    def save(
        self, owner: JidStr, jid: JidStr, state: RosterState, db_state: dict[str, Any]
    ) -> None:
        pass

    def load(
        self, owner: JidStr, jid: JidStr, db_state: dict[str, Any]
    ) -> RosterState | None:
        session = self.xmpp.get_session_from_jid(JID(jid))
        if session is None:
            return {
                "name": "",
                "groups": [],
                "from": False,
                "to": False,
                "pending_in": False,
                "pending_out": False,
                "whitelisted": False,
                "subscription": "both",
            }
        else:
            return {
                "name": "",
                "groups": [],
                "from": True,
                "to": True,
                "pending_in": False,
                "pending_out": False,
                "whitelisted": False,
                "subscription": "none",
            }


log = logging.getLogger(__name__)



"""
Typing stuff
"""

import contextlib
import re
import warnings
from collections.abc import AsyncIterator, Iterable
from dataclasses import dataclass, fields
from datetime import datetime
from enum import IntEnum
from functools import cached_property
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Literal,
    NamedTuple,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
)

import aiohttp
from slixmpp import Message, Presence
from slixmpp.types import PresenceShows, PresenceTypes, ResourceDict  # noqa: F401

if TYPE_CHECKING:
    from ..contact import LegacyContact, LegacyRoster
    from ..core.gateway import BaseGateway
    from ..core.session import BaseSession
    from ..group import LegacyBookmarks, LegacyMUC
    from ..group.participant import LegacyParticipant

AnySession: TypeAlias = "BaseSession[Any]"
AnyGateway: TypeAlias = "BaseGateway[AnySession]"
AnyMUC: TypeAlias = "LegacyMUC[Any]"
AnyBookmarks: TypeAlias = "LegacyBookmarks[Any]"
AnyRoster: TypeAlias = "LegacyRoster[Any]"
AnyParticipant: TypeAlias = "LegacyParticipant[Any]"

LegacyContactType = TypeVar("LegacyContactType", bound="LegacyContact")
LegacyMUCType = TypeVar("LegacyMUCType", bound=AnyMUC)
LegacyParticipantType = TypeVar("LegacyParticipantType", bound=AnyParticipant)

SessionType = TypeVar("SessionType", bound=AnySession)
AnyRecipient = Union["LegacyContact", AnyMUC]
RecipientType = TypeVar("RecipientType", bound=AnyRecipient)
Sender = Union["LegacyContact", "AnyParticipant"]

ChatState = Literal["active", "composing", "gone", "inactive", "paused"]
ProcessingHint = Literal["no-store", "markable", "store"]
Marker = Literal["acknowledged", "received", "displayed"]
FieldType = Literal[
    "boolean",
    "fixed",
    "text-single",
    "text-multi",
    "jid-single",
    "jid-multi",
    "list-single",
    "list-multi",
    "text-private",
]
MucAffiliation = Literal["owner", "admin", "member", "outcast", "none"]
MucRole = Literal["visitor", "participant", "moderator", "none"]
# https://xmpp.org/registrar/disco-categories.html#client
ClientType = Literal[
    "bot", "console", "game", "handheld", "pc", "phone", "sms", "tablet", "web"
]
AttachmentDisposition = Literal["attachment", "inline"]


@dataclass
class MessageReference:
    """
    A "message reply", ie a "quoted message" (:xep:`0461`)

    At the very minimum, the legacy message ID attribute must be set, but to
    ensure that the quote is displayed in all XMPP clients, the author must also
    be set (use the string "user" if the slidge user is the author of the referenced
    message).
    The body is used as a fallback for XMPP clients that do not support :xep:`0461`
    of that failed to find the referenced message.
    """

    legacy_id: str
    author: Union[Literal["user"], AnyParticipant, "LegacyContact"] | None = None
    body: str | None = None


@dataclass
class LegacyAttachment:
    """
    A file attachment to a message

    At the minimum, one of the ``path``, ``steam``, ``data`` or ``url`` attribute
    has to be set

    To be used with :meth:`.LegacyContact.send_files` or
    :meth:`.LegacyParticipant.send_files`
    """

    path: Path | str | None = None
    name: str | None = None
    stream: IO[bytes] | None = None
    aio_stream: AsyncIterator[bytes] | None = None
    data: bytes | None = None
    content_type: str | None = None
    legacy_file_id: str | None = None
    url: str | None = None
    caption: str | None = None
    disposition: AttachmentDisposition | None = None
    is_sticker: bool = False
    """
    A caption for this specific image. For a global caption for a list of attachments,
    use the ``body`` parameter of :meth:`.AttachmentMixin.send_files`
    """

    def __post_init__(self) -> None:
        if all(
            x is None
            for x in (self.path, self.stream, self.data, self.url, self.aio_stream)
        ):
            raise TypeError("There is not data in this attachment", self)
        if isinstance(self.path, str):
            self.path = Path(self.path)
        if self.is_sticker:
            if self.disposition == "attachment":
                warnings.warn(
                    "Sticker declared as 'attachment' disposition, changing it to 'inline'"
                )
            self.disposition = "inline"

    def format_for_user(self) -> str:
        if self.name:
            name = self.name
        elif self.path:
            name = self.path.name  # type:ignore[union-attr]
        elif self.url:
            name = self.url
        else:
            name = ""

        if self.caption:
            name = f"{name}: {self.caption}" if name else self.caption

        return name

    def __str__(self) -> str:
        attrs = ", ".join(
            f"{f.name}={getattr(self, f.name)!r}"
            for f in fields(self)
            if getattr(self, f.name) is not None and f.name != "data"
        )
        if self.data is not None:
            data_str = f"data=<{len(self.data)} bytes>"
            to_join = (attrs, data_str) if attrs else (data_str,)
            attrs = ", ".join(to_join)
        return f"Attachment({attrs})"


class MucType(IntEnum):
    """
    The type of group, private, public, anonymous or not.
    """

    GROUP = 0
    """
    A private group, members-only and non-anonymous, eg a family group.
    """
    CHANNEL = 1
    """
    A public group, aka an anonymous channel.
    """
    CHANNEL_NON_ANONYMOUS = 2
    """
    A public group where participants' legacy IDs are visible to everybody.
    """


PseudoPresenceShow = PresenceShows | Literal[""]


MessageOrPresenceTypeVar = TypeVar("MessageOrPresenceTypeVar", bound=Message | Presence)


class LinkPreview(NamedTuple):
    """
    Embedded metadata from :xep:`0511`.

    See <https://ogp.me/>_.
    """

    about: str
    """
    URL of the link.
    """
    title: str | None
    """
    Title of the linked page.
    """
    description: str | None
    """
    A description of the page.
    """
    url: str | None
    """
    The canonical URL of the link.
    """
    image: str | Path | bytes | None
    """
    An image representing the link. If it is a string, it should represent a URL to an image.
    """
    type: str | None
    """
    Type of the link destination.
    """
    site_name: str | None
    """
    Name of the web site.
    """

    @property
    def is_empty(self) -> bool:
        return not any(x for x in self)


class Mention(NamedTuple):
    contact: "LegacyContact"
    start: int
    end: int


class Hat(NamedTuple):
    uri: str
    title: str
    hue: float | None = None


class UserPreferences(TypedDict):
    sync_avatar: bool
    sync_presence: bool


class MamMetadata(NamedTuple):
    id: str
    sent_on: datetime


class HoleBound(NamedTuple):
    id: str
    timestamp: datetime


class CachedPresence(NamedTuple):
    last_seen: datetime | None = None
    ptype: PresenceTypes | None = None
    pstatus: str | None = None
    pshow: PresenceShows | None = None


class Avatar(NamedTuple):
    path: Path | None = None
    unique_id: str | None = None
    url: str | None = None
    data: bytes | None = None


class SpaceMetadata(NamedTuple):
    creator_legacy_id: str | None = None
    name: str | None = None
    description: str | None = None
    member_count: int | None = None
    owner_legacy_ids: Iterable[str] = []


@dataclass
class Reply:
    msg_id: str
    to: "LegacyContact | LegacyParticipant[Any] | Literal['user'] | None"
    fallback: str | None = None

    @cached_property
    def fallback_no_quote_mark(self) -> str | None:
        """
        Return multi-line text without leading quote marks (i.e. the ">" character).
        """
        if not self.fallback:
            return None
        return re.sub(_STRIP_QUOTE_RE, "", self.fallback).strip()


_STRIP_QUOTE_RE = re.compile(r"^>\s*", flags=re.MULTILINE)


@dataclass
class XMPPAttachment:
    url: str
    is_sticker: bool = False
    cid: str | None = None
    content_type: str | None = None

    @contextlib.asynccontextmanager
    async def get(self) -> AsyncIterator[aiohttp.ClientResponse]:
        async with (
            aiohttp.ClientSession() as session,
            session.get(self.url) as response,
        ):
            yield response


@dataclass
class _AbstractXMPPMessage:
    body: str | None = None
    link_previews: tuple[LinkPreview, ...] = ()
    attachments: tuple[XMPPAttachment, ...] = ()
    mentions: tuple[Mention, ...] = ()
    replace: str | None = None
    reply: Reply | None = None
    thread: str | None = None


class XMPPAttachmentMessage(_AbstractXMPPMessage):
    body: str | None
    attachments: tuple[XMPPAttachment, *tuple[XMPPAttachment, ...]]


class XMPPTextMessage(_AbstractXMPPMessage):
    body: str
    attachments: tuple[()]


XMPPMessage = XMPPAttachmentMessage | XMPPTextMessage


@dataclass
class Sticker:
    path: Path
    content_type: str | None
    hashes: dict[str, str]
    fallback: str | None = None
    reply: Reply | None = None
    thread: str | None = None



from collections.abc import AsyncIterator

from slixmpp.exceptions import XMPPError

from slidge.contact import LegacyContact, LegacyRoster
from slidge.util.types import Avatar, XMPPMessage

from .session import Session


class Roster(LegacyRoster["Contact"]):
    async def fill(self) -> AsyncIterator["Contact"]:
        for i in "111", "222":
            yield await self.by_legacy_id(i)


class Contact(LegacyContact):
    session: "Session"

    @property
    def super_duper_id(self) -> int:
        try:
            return int(self.legacy_id)
        except ValueError:
            raise XMPPError(
                "bad-request", "This is not a valid username for this fake network"
            )

    async def update_info(self) -> None:
        profile = await self.session.legacy_client.get_profile(self.super_duper_id)
        self.name = profile.nickname
        self.set_vcard(full_name=profile.full_name)
        await self.set_avatar(
            Avatar(path=profile.avatar, unique_id=profile.avatar_unique_id)
        )
        if self.legacy_id != "666":
            self.is_friend = True
        self.online()

    async def on_message(self, message: XMPPMessage) -> str | None:
        if message.body:
            sent = await self.session.legacy_client.send_direct_msg(
                message.body, self.super_duper_id
            )
            return sent.id
        return None



from slixmpp.jid import unescape_node


class EscapeMixin:
    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        """
        Convert a legacy ID to a valid 'user' part of a JID.

        The default implementation uses :xep:`0106`.

        Should be overridden for cases where the str conversion of
        the legacy_id is not enough, e.g., if it is case-sensitive or contains
        forbidden characters not covered by :xep:`0106`.

        :param legacy_id: The identifier to convert.
        :return: A valid username part (or "local part") of a JID.
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, jid_username: str) -> str:
        """
        Convert a JID user part to a legacy ID.

        The default implementation uses :xep:`0106`.

        Should be overridden in case legacy IDs are not strings, or more
        generally for any case where the username part of a JID
        is not enough to identify a contact on the legacy network.

        :param jid_username: User part of a JID, ie "user" in "user@example.com"
        :return: The string representation of an identifier on the legacy network.
        """
        return unescape_node(jid_username)  # type:ignore[no-any-return]


JID_ESCAPE_SEQUENCES = {
    "\\20",
    "\\22",
    "\\26",
    "\\27",
    "\\2f",
    "\\3a",
    "\\3c",
    "\\3e",
    "\\40",
    "\\5c",
}


JID_UNESCAPE_TRANSFORMATIONS = {
    "\\20": " ",
    "\\22": '"',
    "\\26": "&",
    "\\27": "'",
    "\\2f": "/",
    "\\3a": ":",
    "\\3c": "<",
    "\\3e": ">",
    "\\40": "@",
    "\\5c": "\\",
}


ESCAPE_TABLE = "".maketrans({v: k for k, v in JID_UNESCAPE_TRANSFORMATIONS.items()})

__all__ = "ESCAPE_TABLE", "unescape_node"



import datetime

from conftest import UUIDFixtureMixin
from slixmpp import Message, Presence
from slixmpp.exceptions import XMPPError

from slidge import (
    BaseGateway,
    BaseSession,
    LegacyBookmarks,
    LegacyContact,
    LegacyParticipant,
    LegacyRoster,
)
from slidge.group import LegacyMUC
from slidge.util.test import SlidgeTest


class Session(BaseSession):
    async def login(self):
        pass


class Gateway(BaseGateway[Session]):
    COMPONENT_NAME = "Megacorp Chat XMPP Gateway"
    GROUPS = True


class Roster(LegacyRoster[LegacyContact]):
    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        if legacy_id.startswith("room-"):
            raise XMPPError("item-not-found", "This is a MUC JID")
        return legacy_id


class MUC(LegacyMUC[LegacyParticipant]):
    async def update_info(self):
        self.subject = "a subject"
        self.subject_setter = "occ1"
        self.subject_date = datetime.datetime(2000, 1, 1)

    async def fill_participants(self):
        for i in range(1, 3):
            yield await self.get_participant_by_legacy_id(f"occ{i}")


class Bookmarks(LegacyBookmarks[MUC]):
    async def fill(self) -> None:
        pass


class TestAffiliationVersioning(UUIDFixtureMixin, SlidgeTest):
    plugin = globals()

    AFF_VER_BASE = "a58097ba3a0ae6791478dac31ae201e0ed2aa06c4aed47fe50ac3b5dc0ea41ce"
    AFF_VER_WITH_OCC3 = (
        "de90fe800d186bfe471e78643829ee5f44b5fdc9bdf55338eacff7a41fb2ed3b"
    )
    AFF_VER_WITHOUT_OCC2 = (
        "55e1628fe7e71daf123613fa594acd86006ec5857533abcfc44b44a73af8f0ef"
    )

    def setUp(self) -> None:
        super().setUp()
        self._add_slidge_user()

    def test_join_no_ver(self) -> None:
        self.__recv_join()
        self.__send_all_affiations()
        self.__send_presences()
        self.__send_subject()
        self.__send_vcard()

    def test_join_no_diff(self) -> None:
        self.__recv_join(since=self.AFF_VER_BASE)
        self.__send_presences()
        self.__send_subject()
        self.__send_vcard()

    def test_join_unknown_ver(self) -> None:
        self.__recv_join(since="unkown")
        self.__send_all_affiations()
        self.__send_presences()
        self.__send_subject()
        self.__send_vcard()

    def test_live_new_member(self) -> None:
        self.test_join_no_ver()
        muc: LegacyMUC = self.run_coro(
            self.user_session().bookmarks.by_legacy_id("room-whatever")
        )
        occ3 = self.run_coro(muc.get_participant_by_legacy_id("occ3"))
        presence = self.next_sent()
        assert isinstance(presence, Presence)
        assert presence.get_from() == occ3.jid
        self.send(  # language=XML
            f"""
            <message from='room-whatever@aim.shakespeare.lit'
                     to='romeo@shakespeare.lit/desktop'
                     type='normal'>
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="3"
                         by="room-whatever@aim.shakespeare.lit" />
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <mav xmlns='urn:xmpp:muc:affiliations:1'
                     since='{self.AFF_VER_BASE}'
                     until='{self.AFF_VER_WITH_OCC3}' />
                <item jid='occ3@aim.shakespeare.lit'
                      affiliation='member' />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="room" />
            </message>
            """,
            use_values=False,
        )

    def test_live_revoke(self) -> None:
        self.test_join_no_ver()
        muc: MUC = self.run_coro(
            self.user_session().bookmarks.by_legacy_id("room-whatever")
        )
        occ2: LegacyParticipant = self.run_coro(
            muc.get_participant_by_legacy_id("occ2")
        )
        occ2.leave()
        self.send(  # language=XML
            f"""
            <message from='room-whatever@aim.shakespeare.lit'
                     to='romeo@shakespeare.lit/desktop'
                     type='normal'>
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="3"
                         by="room-whatever@aim.shakespeare.lit" />
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <mav xmlns='urn:xmpp:muc:affiliations:1'
                     since='{self.AFF_VER_BASE}'
                     until='{self.AFF_VER_WITHOUT_OCC2}' />
                <item jid='occ2@aim.shakespeare.lit'
                      affiliation='none' />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="room" />
            </message>
            """,
            use_values=False,
        )
        presence = self.next_sent()
        assert isinstance(presence, Presence)
        assert presence.get_from() == occ2.jid

    def __recv_join(self, since: str = "") -> None:
        if since:
            mav = f"<mav xmlns='urn:xmpp:muc:affiliations:1' since='{since}' />"
        else:
            mav = "<mav xmlns='urn:xmpp:muc:affiliations:1' />"
        self.recv(  # language=XML
            f"""
            <presence to='room-whatever@aim.shakespeare.lit/romeo'
                      from='romeo@shakespeare.lit/desktop'>
              <x xmlns='http://jabber.org/protocol/muc'>{mav}</x>
            </presence>
            """
        )

    def __send_all_affiations(self) -> None:
        self.send(  # language=XML
            f"""
            <message from='room-whatever@aim.shakespeare.lit'
                     to='romeo@shakespeare.lit/desktop'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <mav xmlns="urn:xmpp:muc:affiliations:1"
                     until='{self.AFF_VER_BASE}' />
                <item jid='occ1@aim.shakespeare.lit'
                      affiliation='member' />
                <item jid='occ2@aim.shakespeare.lit'
                      affiliation='member' />
                <item jid='romeo@shakespeare.lit'
                      affiliation='member' />
              </x>
            </message>
            """,
            use_values=False,
        )

    def __send_presences(self) -> None:
        for i in range(1, 3):
            pres = self.next_sent()
            assert isinstance(pres, Presence)
            assert pres.get_from().resource == f"occ{i}", pres
        self.send(  # language=XML
            f"""
            <presence from='room-whatever@aim.shakespeare.lit/romeo'
                      to='romeo@shakespeare.lit/desktop'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <mav xmlns='urn:xmpp:muc:affiliations:1'
                     until='{self.AFF_VER_BASE}' />
                <item affiliation='member'
                      role='participant' />
                <status code='110' />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </presence>
            """,
            use_values=False,
        )

    def __send_subject(self) -> None:
        subject = self.next_sent()
        assert isinstance(subject, Message)
        assert "subject" in subject, subject

    def __send_vcard(self) -> None:
        _vcard = self.next_sent()
        assert self.next_sent() is None



"""Spaces

Revision ID: bdf1c8b85ea3
Revises: e365e04e9ac3
Create Date: 2026-05-19 20:20:35.452386

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "bdf1c8b85ea3"
down_revision: str | None = "e365e04e9ac3"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.create_table(
        "space",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("updated", sa.Boolean(), nullable=False),
        sa.Column("user_account_id", sa.Integer(), nullable=False),
        sa.Column("legacy_id", sa.String(), nullable=False),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("member_count", sa.Integer(), nullable=True),
        sa.Column("creator_pk", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["creator_pk"], ["contact.id"], name=op.f("fk_space_creator_pk_contact")
        ),
        sa.ForeignKeyConstraint(
            ["user_account_id"],
            ["user_account.id"],
            name=op.f("fk_space_user_account_id_user_account"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_space")),
        sa.UniqueConstraint(
            "user_account_id", "legacy_id", name="uq_space_user_account_id_legacy_id"
        ),
    )
    op.create_table(
        "space_owner_association",
        sa.Column("space_id", sa.Integer(), nullable=False),
        sa.Column("contact_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["contact_id"],
            ["contact.id"],
            name=op.f("fk_space_owner_association_contact_id_contact"),
        ),
        sa.ForeignKeyConstraint(
            ["space_id"],
            ["space.id"],
            name=op.f("fk_space_owner_association_space_id_space"),
        ),
        sa.PrimaryKeyConstraint(
            "space_id", "contact_id", name=op.f("pk_space_owner_association")
        ),
    )
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.add_column(sa.Column("space_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            batch_op.f("fk_room_space_id_space"), "space", ["space_id"], ["id"]
        )


def downgrade() -> None:
    with op.batch_alter_table("room", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_room_space_id_space"), type_="foreignkey"
        )
        batch_op.drop_column("space_id")

    op.drop_table("space_owner_association")
    op.drop_table("space")



# Commands available to users
import contextlib
from copy import deepcopy
from typing import Any, cast

from slixmpp import JID
from slixmpp.exceptions import XMPPError

from slidge.db.meta import JSONSerializable
from slidge.db.models import Space

from ..util.types import (
    AnyMUC,
    AnySession,
    MucType,
    UserPreferences,
)
from .base import (
    Command,
    CommandAccess,
    Confirmation,
    ConfirmationSession,
    Form,
    FormField,
    FormSession,
    FormValues,
    SearchResult,
    TableResult,
)
from .categories import CONTACTS, GROUPS, SPACES


class Search(Command[AnySession]):
    NAME = "🔎 Search for contacts"
    HELP = "Search for contacts via this gateway"
    CHAT_COMMAND = "find"
    NODE = CONTACTS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = CONTACTS

    async def run(
        self, session: "AnySession | None", _ifrom: JID, *args: str
    ) -> FormSession[AnySession] | SearchResult | None:
        assert session is not None
        await session.ready
        if args:
            return await session.on_search(
                {self.xmpp.SEARCH_FIELDS[0].var: " ".join(args)}
            )
        return FormSession[AnySession](
            title=self.xmpp.SEARCH_TITLE,
            instructions=self.xmpp.SEARCH_INSTRUCTIONS,
            fields=self.xmpp.SEARCH_FIELDS,
            handler=self.search,
        )

    @staticmethod
    async def search(
        form_values: FormValues, session: "AnySession | None", _ifrom: JID
    ) -> SearchResult:
        assert session is not None
        results = await session.on_search(form_values)  # type: ignore
        if results is None:
            raise XMPPError("item-not-found", "No contact was found")

        return results


class SyncContacts(Command[AnySession]):
    NAME = "🔄 Sync XMPP roster"
    HELP = (
        "Synchronize your XMPP roster with your legacy contacts. "
        "Slidge will only add/remove/modify contacts in its dedicated roster group"
    )
    CHAT_COMMAND = "sync-contacts"
    NODE = CONTACTS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = CONTACTS

    async def run(
        self,
        session: "AnySession | None",
        _ifrom: JID,
        *args: str,
    ) -> ConfirmationSession[AnySession]:
        assert session is not None
        await session.ready
        return ConfirmationSession[AnySession](
            prompt="Are you sure you want to sync your roster?",
            success=None,
            handler=self.sync,
        )

    async def sync(self, session: "AnySession | None", _ifrom: JID) -> str:
        if session is None:
            raise RuntimeError
        roster_iq = await self.xmpp["xep_0356"].get_roster(session.user_jid.bare)

        contacts = session.contacts.known_contacts()

        added = 0
        removed = 0
        updated = 0
        for item in roster_iq["roster"]:
            groups = set(item["groups"])
            if self.xmpp.ROSTER_GROUP in groups:
                contact = contacts.pop(item["jid"], None)
                if contact is None:
                    if len(groups) == 1:
                        await self.xmpp["xep_0356"].set_roster(
                            session.user_jid, {item["jid"]: {"subscription": "remove"}}
                        )
                        removed += 1
                    else:
                        groups.remove(self.xmpp.ROSTER_GROUP)
                        await self.xmpp["xep_0356"].set_roster(
                            session.user_jid,
                            {
                                item["jid"]: {
                                    "subscription": item["subscription"],
                                    "name": item["name"],
                                    "groups": groups,
                                }
                            },
                        )
                        updated += 1
                else:
                    if contact.name != item["name"]:
                        await contact.add_to_roster(force=True)
                        updated += 1

        # we popped before so this only acts on slidge contacts not in the xmpp roster
        for contact in contacts.values():
            added += 1
            await contact.add_to_roster()

        return f"{added} added, {removed} removed, {updated} updated"


class ListContacts(Command[AnySession]):
    NAME = HELP = "👤 List your legacy contacts"
    CHAT_COMMAND = "contacts"
    NODE = CONTACTS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = CONTACTS

    async def run(
        self,
        session: "AnySession | None",
        _ifrom: JID,
        *_: str,
    ) -> TableResult:
        assert session is not None
        await session.ready
        contacts = sorted(
            session.contacts, key=lambda c: c.name.casefold() if c.name else ""
        )
        return TableResult(
            description="Your buddies",
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            items=[{"name": c.name, "jid": c.jid.bare} for c in contacts],
        )


class ListGroups(Command[AnySession]):
    NAME = HELP = "👥 List your legacy groups"
    CHAT_COMMAND = "groups"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = GROUPS

    async def run(
        self, session: "AnySession | None", _ifrom: JID, *_: str
    ) -> TableResult:
        assert session is not None
        await session.ready
        groups: list[AnyMUC] = sorted(
            session.bookmarks, key=lambda g: (g.name or g.jid.node).casefold()
        )
        return TableResult(
            description="Your groups",
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            items=[
                {"name": g.name or str(g.legacy_id), "jid": g.jid.bare} for g in groups
            ],
            jids_are_mucs=True,
        )


class ListSpaces(Command[AnySession]):
    NAME = "🌐 List my spaces"
    HELP = "List the spaces you are part of. Spaces are collections of rooms."
    CHAT_COMMAND = "spaces"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = SPACES

    related_to_spaces = True

    async def run(
        self, session: "AnySession | None", _ifrom: JID, *_: str
    ) -> TableResult:
        assert session is not None
        spaces = await _get_updated_spaces(session)
        return TableResult(
            description="Your spaces. If your client does not support spaces, use the 'space-rooms' command.",
            fields=[FormField("name"), FormField("iri")],
            items=[
                {
                    "name": s.name or str(s.legacy_id),
                    "iri": f"xmpp:{self.xmpp.boundjid.bare}?;node={await session.bookmarks.space_legacy_id_to_node(s.legacy_id)}",
                }
                for s in spaces
            ],
            jids_are_mucs=True,
        )


class ListRoomsInSpace(Command[AnySession]):
    NAME = "🌐 List the rooms in a space"
    HELP = "List the rooms of a space you are part of. Spaces are collections of rooms."
    CHAT_COMMAND = "space-rooms"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = SPACES

    async def run(
        self, session: "AnySession | None", _ifrom: JID, *_: str
    ) -> "FormSession[AnySession]":
        assert session is not None
        spaces = await _get_updated_spaces(session)
        return FormSession(
            title="Your spaces",
            instructions="Select a space to view its rooms",
            fields=[
                FormField(
                    var="space_legacy_id",
                    label="Space",
                    required=True,
                    type="list-single",
                    options=[
                        {
                            "label": s.name or str(s.legacy_id),
                            "value": str(s.legacy_id),
                        }
                        for s in spaces
                    ],
                )
            ],
            handler=self.list_rooms,
            handler_args=({str(s.legacy_id): s.name for s in spaces},),
        )

    async def list_rooms(
        self,
        form_values: FormValues,
        session: "AnySession | None",
        _ifrom: JID,
        space_names: dict[str, str],
    ) -> TableResult:
        assert session is not None
        space_legacy_id = form_values.get("space_legacy_id")
        if space_legacy_id is None:
            raise XMPPError("bad-request", "You need to specify a space")
        assert isinstance(space_legacy_id, str)
        await session.ready
        with self.xmpp.store.session() as orm:
            rooms = sorted(
                list(
                    self.xmpp.store.spaces.get_rooms(
                        orm,
                        session.user_pk,
                        await session.bookmarks.space_legacy_id_to_node(
                            space_legacy_id
                        ),
                    )
                ),
                key=lambda r: (r.name or str(r.jid.node)).casefold(),
            )
        name = space_names.get(space_legacy_id)
        return TableResult(
            fields=[
                FormField("name", "Name"),
                FormField("jid", "JID", type="jid-single"),
            ],
            description=f"Rooms of '{name or space_legacy_id}'",
            items=[{"name": r.name or str(r.legacy_id), "jid": r.jid} for r in rooms],
            jids_are_mucs=True,
        )


class Login(Command[AnySession]):
    NAME = "🔐 Re-login to the legacy network"
    HELP = "Login to the legacy service"
    CHAT_COMMAND = "re-login"
    NODE = "https://slidge.im/command/core/" + CHAT_COMMAND

    ACCESS = CommandAccess.USER_NON_LOGGED

    async def run(
        self,
        session: "AnySession | None",
        _ifrom: JID,
        *_: str,
    ) -> str:
        assert session is not None
        if session.is_logging_in:
            raise XMPPError("bad-request", "You are already logging in.")
        session.is_logging_in = True
        try:
            msg = await self.xmpp.login_wrap(session)
        except Exception as e:
            session.send_gateway_status(f"Re-login failed: {e}", show="dnd")
            raise XMPPError(
                "internal-server-error", etype="wait", text=f"Could not login: {e}"
            )
        finally:
            session.is_logging_in = False
        session.logged = True
        session.send_gateway_status(msg or "Re-connected", show="chat")
        session.send_gateway_message(msg or "Re-connected")
        return msg


class CreateGroup(Command[AnySession]):
    NAME = "🆕 New legacy group"
    HELP = "Create a group on the legacy service"
    CHAT_COMMAND = "create-group"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    CATEGORY = GROUPS

    ACCESS = CommandAccess.USER_LOGGED

    async def run(
        self,
        session: "AnySession | None",
        _ifrom: JID,
        *_: str,
    ) -> FormSession[AnySession]:
        assert session is not None
        await session.ready
        contacts = session.contacts.known_contacts(only_friends=True)
        return FormSession(
            title="Create a new group",
            instructions="Pick contacts that should be part of this new group",
            fields=[
                FormField(var="group_name", label="Name of the group", required=True),
                FormField(
                    var="contacts",
                    label="Contacts to add to the new group",
                    type="list-multi",
                    options=[
                        {"value": str(contact.jid), "label": contact.name}
                        for contact in sorted(contacts.values(), key=lambda c: c.name)
                    ],
                    required=False,
                ),
            ],
            handler=self.finish,
        )

    @staticmethod
    async def finish(
        form_values: FormValues,
        session: "AnySession | None",
        *_: Any,  # noqa:ANN401
    ) -> TableResult:
        assert session is not None
        legacy_id: str = await session.on_create_group(
            cast(str, form_values["group_name"]),
            [
                await session.contacts.by_jid(JID(j))
                for j in form_values.get("contacts", [])  # type:ignore
            ],
        )
        muc = await session.bookmarks.by_legacy_id(legacy_id)
        return TableResult(
            description=f"Your new group: xmpp:{muc.jid}?join",
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            items=[{"name": muc.name, "jid": muc.jid}],
            jids_are_mucs=True,
        )


class Preferences(Command[AnySession]):
    NAME = "⚙️ Preferences"
    HELP = "Customize the gateway behaviour to your liking"
    CHAT_COMMAND = "preferences"
    NODE = "https://slidge.im/command/core/preferences"
    ACCESS = CommandAccess.USER

    async def run(
        self,
        session: "AnySession | None",
        _ifrom: JID,
        *_: str,
    ) -> FormSession[AnySession]:
        fields = deepcopy(self.xmpp.PREFERENCES)
        assert session is not None
        current = session.user.preferences
        for field in fields:
            field.value = current.get(field.var, field.value)  # type:ignore
        return Form(
            title="Preferences",
            instructions=self.HELP,
            fields=fields,
            handler=self.finish,  # type:ignore
            handler_kwargs={"previous": current},
        )

    async def finish(
        self,
        form_values: UserPreferences,
        session: "AnySession | None",
        *_: Any,  # noqa:ANN401
        previous: JSONSerializable,
    ) -> str:
        assert session is not None
        if previous == form_values:
            return "No preference was changed"

        previous = previous.copy()
        user = session.user
        user.preferences.update(form_values)  # type:ignore
        self.xmpp.store.users.update(user)

        with contextlib.suppress(NotImplementedError):
            await session.on_preferences(previous, form_values)  # type:ignore[arg-type]

        if not previous["sync_avatar"] and form_values["sync_avatar"]:
            await self.xmpp.fetch_user_avatar(session)
        else:
            user.avatar_hash = None

        return "Your preferences have been updated."


class Unregister(Command[AnySession]):
    NAME = "❌ Unregister from the gateway"
    HELP = "Unregister from the gateway"
    CHAT_COMMAND = "unregister"
    NODE = "https://slidge.im/command/core/unregister"
    ACCESS = CommandAccess.USER

    async def run(
        self, session: "AnySession | None", _ifrom: JID, *_: str
    ) -> ConfirmationSession[AnySession]:
        return ConfirmationSession[AnySession](
            prompt=f"Are you sure you want to unregister from '{self.xmpp.boundjid}'?",
            success=f"You are not registered to '{self.xmpp.boundjid}' anymore.",
            handler=self.unregister,
        )

    async def unregister(self, session: "AnySession | None", _ifrom: JID) -> str:
        assert session is not None
        await self.xmpp.unregister_user(session.user)
        return "You are not registered anymore. Bye!"


class LeaveGroup(Command[AnySession]):
    NAME = HELP = "❌ Leave a legacy group"
    CHAT_COMMAND = "leave-group"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = GROUPS

    async def run(
        self,
        session: "AnySession | None",
        ifrom: JID,
        *_: str,
    ) -> FormSession[AnySession]:
        assert session is not None
        await session.ready
        groups = sorted(session.bookmarks, key=lambda g: g.DISCO_NAME.casefold())
        return FormSession(
            title="Leave a group",
            instructions="Select the group you want to leave",
            fields=[
                FormField(
                    "group",
                    "Group name",
                    type="list-single",
                    options=[
                        {"label": g.name or str(g.legacy_id), "value": str(i)}
                        for i, g in enumerate(groups)
                    ],
                )
            ],
            handler=self.confirm,
            handler_args=(groups,),
        )

    async def confirm(
        self,
        form_values: FormValues,
        _session: "AnySession | None",
        _ifrom: JID,
        groups: list[AnyMUC],
    ) -> Confirmation:
        group = groups[int(form_values["group"])]  # type:ignore
        return Confirmation(
            prompt=f"Are you sure you want to leave the group '{group.name}'?",
            handler=self.finish,
            handler_args=(group,),
        )

    @staticmethod
    async def finish(session: AnySession, _ifrom: JID, group: AnyMUC) -> None:
        await group.on_leave()
        await session.bookmarks.remove(group, reason="You left this group via slidge.")


class InviteInGroups(Command[AnySession]):
    NAME = "💌 Re-invite me in my groups"
    HELP = "Ask the gateway to send invitations for all your private groups"
    CHAT_COMMAND = "re-invite"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = GROUPS

    async def run(self, session: "AnySession | None", _ifrom: JID, *_: str) -> None:
        assert session is not None
        await session.ready
        for muc in session.bookmarks:
            if muc.type == MucType.GROUP:
                session.send_gateway_invite(
                    muc, reason="You asked to be re-invited in all groups."
                )


async def _get_updated_spaces(session: AnySession) -> list[Space]:
    await session.ready
    with session.xmpp.store.session() as orm:
        spaces = list(session.xmpp.store.spaces.get_all(orm, session.user_pk))
    updated_spaces: list[Space] = []
    for space in spaces:
        try:
            updated_spaces.append(await session.bookmarks.update_space_if_needed(space))
        except Exception:
            session.log.exception(
                "Something went wrong trying to update space '%r'", space
            )

    return sorted(updated_spaces, key=lambda s: s.name or str(s.legacy_id))



"""
Handling groups
"""

import uuid
from collections.abc import AsyncIterator
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from slixmpp.exceptions import XMPPError

from slidge.group.bookmarks import LegacyBookmarks
from slidge.group.participant import LegacyParticipant
from slidge.group.room import LegacyMUC
from slidge.util.types import (
    Avatar,
    Hat,
    HoleBound,
    MucType,
    SpaceMetadata,
    XMPPMessage,
)

if TYPE_CHECKING:
    from .session import Session


from .contact import Contact


class Participant(LegacyParticipant[Contact]):
    pass


class Bookmarks(LegacyBookmarks["MUC"]):
    async def fill(self) -> None:
        for i in "aaa", "bbb":
            muc = await self.by_legacy_id(i)
            await muc.add_to_bookmarks()

    async def fetch_space_metadata(self, legacy_id: str) -> SpaceMetadata:
        if legacy_id == "555":
            return SpaceMetadata(
                name="The cool super space",
                creator_legacy_id="111",
                owner_legacy_ids=["111", "666"],
            )
        raise XMPPError("item-not-found", f"This space is not a space: {legacy_id}")


class MUC(LegacyMUC[Participant]):
    session: "Session"
    type = MucType.GROUP

    async def update_info(self) -> None:
        info = await self.session.legacy_client.get_group_info(self.legacy_id)
        self.space_legacy_id = "555"
        self.name = info.name
        await self.set_avatar(Avatar(path=info.avatar, unique_id=info.avatar_unique_id))

    async def fill_participants(self) -> AsyncIterator["Participant"]:
        # in a real case, this would probably call something like
        # self.session.legacy_client.fetch_group_members(self.legacy_id)
        for i in "0", "111", "222":
            part = await self.get_participant_by_legacy_id(i)
            if i == "111":
                part.role = "moderator"
                part.affiliation = "owner"
                part.set_hats([Hat("test", "test"), Hat("prout", "prout")])
            yield part
        me = await self.get_user_participant()
        me.role = "moderator"
        me.affiliation = "owner"
        yield me

    async def backfill(
        self,
        after: HoleBound | None = None,
        before: HoleBound | None = None,
    ) -> None:
        # in a real case, this would probably call something like
        # self.session.legacy_client.fetch_group_history(self.legacy_id)
        for i in range(10):
            part = await self.get_participant_by_legacy_id("0")
            part.send_text(
                f"History message #{i}",
                when=datetime.now() - timedelta(hours=i),
                legacy_msg_id=f"{i}--{uuid.uuid4().hex}",
                archive_only=True,
            )

    async def on_message(self, message: XMPPMessage) -> str | None:
        if message.body:
            return (
                await self.session.legacy_client.send_group_msg(
                    message.body, self.legacy_id
                )
            ).id
        return None
